#!/usr/bin/env python
"""
Base Module for Testing GUI
 - responsible to set up Qt in order to run on multiple platforms
 - responsible to set up Application
class TestBase
    - responsible to set up and tearDown test workspace
    - responsible to create easy access to 'resource' directory
    - responsible to set up ErrorDialog for easy evaluation of raised exceptions
    - provide method `manual_use` created to help test development process
class DragAndDrop
    - responsible to perform Drag and Drop operations
     from source widget - specific point, to destination widget - specific point
"""

from collections.abc import Iterable
import functools
import os
import pathlib
import shutil
import sys
import time
from typing import Literal
import unittest
from unittest import mock

from h5py import File as HDF5
import numpy as np
from numpy.typing import NDArray
import pyqtgraph
from PySide6 import QtCore, QtGui, QtTest, QtWidgets

from asammdf import mdf
from asammdf.gui.utils import excepthook
from asammdf.gui.widgets.tree import ChannelsTreeItem

if sys.platform == "win32":
    os.environ["QT_QPA_PLATFORM"] = "windows"
elif sys.platform == "linux":
    os.environ["QT_QPA_PLATFORM"] = "xcb"
elif sys.platform == "darwin":
    os.environ["QT_QPA_PLATFORM"] = "cocoa"
else:
    os.environ["QT_QPA_PLATFORM"] = "windows"

app = pyqtgraph.mkQApp()
app.setOrganizationName("py-asammdf")
app.setOrganizationDomain("py-asammdf")
app.setApplicationName("py-asammdf")


def safe_setup(func):
    """Decorator for a TestCase setUp method.
    Ensures that if setUp raises an exception, tearDown is still called.
    """

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        except:
            if hasattr(self, "tearDown"):
                self.tearDown()
            raise

    return wrapper


@unittest.skipIf(sys.platform == "darwin", "Test Development on MacOS was not done yet.")
class TestBase(unittest.TestCase):
    """
    - setUp and tearDown test workspace
    - provide method to execute widget
    - setUp ErrorDialog for evaluation raised exceptions
    """

    pyqtgraph.setConfigOption("background", "k")
    pyqtgraph.setConfigOption("foreground", "w")

    settings = QtCore.QSettings()
    settings.setValue("plot/zoom/x_center_on_cursor", True)
    settings.setValue("plot/cursor/display_precision", 6)

    longMessage = False

    resource = os.path.normpath(os.path.join(os.path.dirname(__file__), "resources"))
    test_workspace = os.path.join(os.path.dirname(__file__), "test_workspace")
    screenshots = os.path.join(os.path.dirname(__file__), "screenshots")

    patchers = []
    # MockClass ErrorDialog
    mc_ErrorDialog = None

    def shortDescription(self):
        return self._testMethodDoc

    @staticmethod
    def manual_use(w, duration=None):
        """
        Execute Widget for debug/development purpose.

        Parameters
        ----------
        duration : float | None
            duration in seconds
        """
        if duration is None:
            duration = 3600
        else:
            duration = abs(duration)

        w.showMaximized()

        loop = QtCore.QEventLoop()
        QtCore.QTimer.singleShot(int(duration * 1000), loop.quit)
        loop.exec()

        w.showMaximized()

    @staticmethod
    def processEvents(timeout=0.001):
        t_end = time.perf_counter() + timeout
        while True:
            time.sleep(0.001)
            QtCore.QCoreApplication.processEvents(flags=QtCore.QEventLoop.ProcessEventsFlag.AllEvents)
            if time.perf_counter() > t_end:
                break

    @safe_setup
    def setUp(self) -> None:
        if os.path.exists(self.test_workspace):
            try:
                shutil.rmtree(self.test_workspace)
            except PermissionError as e:
                print(e)

        os.makedirs(self.screenshots, exist_ok=True)
        os.makedirs(self.test_workspace, exist_ok=True)

        self.mc_ErrorDialog.reset_mock()
        self.processEvents()

    @classmethod
    def setUpClass(cls):
        sys.excepthook = excepthook
        for patcher in (mock.patch("asammdf.gui.utils.ErrorDialog"),):
            _ = patcher.start()
            cls.patchers.append(patcher)
        cls.mc_ErrorDialog = _

    @classmethod
    def tearDownClass(cls):
        for patcher in cls.patchers:
            patcher.stop()

    def tearDown(self):
        self.processEvents()
        w = getattr(self, "widget", None)
        if w:
            self.take_screenshot(w)
            self.destroy(w)

        self.mc_ErrorDialog.reset_mock()

        if self.test_workspace and pathlib.Path(self.test_workspace).exists():
            shutil.rmtree(self.test_workspace)

    def take_screenshot(self, widget):
        path = self.screenshots
        for name in self.id().split(".")[:-1]:
            _path = os.path.join(path, name)
            if not os.path.exists(_path):
                os.makedirs(_path)
            path = _path
        widget.grab().save(os.path.join(path, f"{self.id().split('.')[-1]}.png"))

    @staticmethod
    def destroy(w):
        w.close()
        w.destroy()
        w.deleteLater()

    def mouseClick_RadioButton(self, qitem):
        QtTest.QTest.mouseClick(
            qitem,
            QtCore.Qt.MouseButton.LeftButton,
            QtCore.Qt.KeyboardModifier.NoModifier,
            QtCore.QPoint(2, qitem.height() / 2),
        )
        self.processEvents()

    def mouseClick_CheckboxButton(self, qitem):
        # Same function
        self.mouseClick_RadioButton(qitem)

    def mouseClick_WidgetItem(self, qitem):
        if isinstance(qitem, QtWidgets.QTreeWidgetItem):
            widget = qitem.treeWidget()
        elif isinstance(qitem, QtWidgets.QListWidgetItem):
            widget = qitem.listWidget()
        else:
            raise NotImplementedError
        QtTest.QTest.mouseClick(
            widget.viewport(),
            QtCore.Qt.MouseButton.LeftButton,
            QtCore.Qt.KeyboardModifier.NoModifier,
            widget.visualItemRect(qitem).center(),
        )
        self.processEvents(0.5)

    def mouseDClick_WidgetItem(self, qitem):
        if isinstance(qitem, QtWidgets.QTreeWidgetItem):
            widget = qitem.treeWidget()
        elif isinstance(qitem, QtWidgets.QListWidgetItem):
            widget = qitem.listWidget()
        else:
            raise NotImplementedError
        QtTest.QTest.mouseDClick(
            widget.viewport(),
            QtCore.Qt.MouseButton.LeftButton,
            QtCore.Qt.KeyboardModifier.NoModifier,
            widget.visualItemRect(qitem).center(),
        )
        self.processEvents(0.5)

    def is_not_blinking(self, to_grab: QtWidgets.QWidget, colors: set[str], timeout: float = 5.0) -> bool:
        """
        Parameters
        ----------
        - to_grab: widget ex: self.plot
        - colors: a set of colors names ex: {"#123456", "#ffffff"}
        Returns
        -------
        - True: if not timeout and all colors exist on pixmap
        - False: if timeout and no colors exist on pixmap
        """
        if not hasattr(to_grab, "grab"):
            raise Warning(f"object {to_grab} has no attribute grab")

        now = time.perf_counter()
        while True:
            self.processEvents(0.01)
            all_colors = Pixmap.color_names_exclude_defaults(to_grab.grab())
            if colors.issubset(all_colors):
                break
            if time.perf_counter() - now > timeout:
                return False
        return True


class DragAndDrop:
    def __init__(self, src_widget, dst_widget, src_pos, dst_pos):
        QtCore.QCoreApplication.processEvents()
        # hack QDrag object
        with mock.patch(f"{src_widget.__module__}.QtGui.QDrag") as mo_QDrag:
            src_widget.startDrag(QtCore.Qt.DropAction.MoveAction)
            mo_QDrag.assert_called()
            mime_data = mo_QDrag.return_value.setMimeData.call_args.args[0]

            event = QtGui.QDragEnterEvent(
                dst_pos,
                QtCore.Qt.DropAction.MoveAction,
                mime_data,
                QtCore.Qt.MouseButton.LeftButton,
                QtCore.Qt.KeyboardModifier.NoModifier,
            )
            with mock.patch.object(event, "source", return_value=src_widget):
                dst_widget.dragEnterEvent(event)

            event = QtGui.QDropEvent(
                dst_pos,
                QtCore.Qt.DropAction.MoveAction,
                mime_data,
                QtCore.Qt.MouseButton.LeftButton,
                QtCore.Qt.KeyboardModifier.NoModifier,
            )
            with mock.patch.object(event, "source", return_value=src_widget):
                dst_widget.dropEvent(event)
        QtCore.QCoreApplication.processEvents()


class Pixmap:
    COLOR_BACKGROUND = "#000000"
    COLOR_RANGE = "#000032"
    COLOR_CURSOR = "#e69138"

    @staticmethod
    def to_numpy(pixmap: QtGui.QPixmap) -> NDArray[np.uint32]:
        """
        Convert a QPixmap to a NumPy array of 32-bit ARGB values.

        This function returns a new NumPy array with a copy of the pixel data, so
        modifications to the array do not affect the original QPixmap.

        Parameters
        ----------
        pixmap : QtGui.QPixmap
            The QPixmap to convert.

        Returns
        -------
        NDArray[np.uint32]
            A 2D NumPy array of shape (height, width), where each element is a
            32-bit unsigned integer representing a pixel in ARGB format.
        """
        image = pixmap.toImage().convertToFormat(QtGui.QImage.Format.Format_ARGB32)
        ptr = image.bits()
        arr = np.frombuffer(ptr, np.uint32).reshape((pixmap.height(), pixmap.width()))
        return arr.copy()

    @staticmethod
    def is_black(pixmap: QtGui.QPixmap) -> bool:
        """
        Check if a QPixmap is entirely black, ignoring background and cursor colors.

        The function computes all unique colors in the pixmap, excludes the
        predefined background and cursor colors, and determines whether any
        remaining colors exist.

        Parameters
        ----------
        pixmap : QtGui.QPixmap
            The QPixmap to check.

        Returns
        -------
        bool
            True if all visible pixels are black (excluding background and cursor),
            False otherwise.
        """
        ignored_colors = {Pixmap.COLOR_BACKGROUND, Pixmap.COLOR_CURSOR}
        colors = Pixmap.color_names(pixmap) - ignored_colors
        return len(colors) == 0

    @staticmethod
    def is_colored(
        pixmap: QtGui.QPixmap,
        color_name: str,
        x: int,
        y: int,
        width: int | None = None,
        height: int | None = None,
    ) -> bool:
        """
        Check whether a rectangular region of a QPixmap is entirely filled with a given color.

        Parameters
        ----------
        pixmap : QtGui.QPixmap
            The QPixmap to inspect.
        color_name : str
            The target color, expressed as a hex string (e.g., "#ff0000").
        x : int
            The x-coordinate of the top-left corner of the rectangle.
        y : int
            The y-coordinate of the top-left corner of the rectangle.
        width : int | None, optional
            The width of the rectangle. If None, defaults to the width of the pixmap from x.
        height : int | None, optional
            The height of the rectangle. If None, defaults to the height of the pixmap from y.

        Returns
        -------
        bool
            True if all pixels in the specified rectangle have the given color, False otherwise.

        """
        width = width or (pixmap.width() - x)
        height = height or (pixmap.height() - x)

        arr = Pixmap.to_numpy(pixmap)
        arr_rect = arr[y : y + height, x : x + width]
        colors_in_rect = {QtGui.QColor.fromRgba(x).name() for x in np.unique_values(arr_rect)}
        return colors_in_rect == {color_name}

    @staticmethod
    def has_color(pixmap: QtGui.QPixmap, color_name: str | QtGui.QColor | ChannelsTreeItem, tolerance: int = 0) -> bool:
        """
        Return True if "pixmap" contains a color close to "color_name", allowing some tolerance.

        This function accounts for anti-aliasing and smoothing by checking
        if any pixel's color is within a specified tolerance of the target color.

        Parameters
        ----------
        pixmap : QtGui.QPixmap
            The pixmap to check.
        color_name : QtGui.QColor or str
            The color to detect, either as a QColor or hex string (e.g., "#ff0000").
        tolerance : int
            Maximum allowed Euclidean distance in RGB space to consider a pixel as matching.

        Returns
        -------
        bool
            True if at least one pixel in the pixmap is close enough to the target color.
        """
        if isinstance(color_name, str):
            color_name = QtGui.QColor(color_name)
        elif isinstance(color_name, ChannelsTreeItem):
            color_name = QtGui.QColor(color_name.color)

        arr = Pixmap.to_numpy(pixmap)
        # Extract RGB components
        r = (arr >> 16) & 0xFF
        g = (arr >> 8) & 0xFF
        b = arr & 0xFF

        # Compute squared distance to target color in RGB space
        dr = r.astype(np.int64) - color_name.red()
        dg = g.astype(np.int64) - color_name.green()
        db = b.astype(np.int64) - color_name.blue()
        dist_sq = np.square(dr) + np.square(dg) + np.square(db)

        return bool(np.any(dist_sq <= tolerance**2))

    @staticmethod
    def color_names(pixmap: QtGui.QPixmap) -> set[str]:
        """
        Return a set of all unique colors in a QPixmap as hex strings.

        The function converts the QPixmap to a NumPy array of 32-bit ARGB values,
        finds all unique pixel values, and converts each to a standard hex color
        string (e.g., "#ff0000").

        Parameters
        ----------
        pixmap : QtGui.QPixmap
            The QPixmap from which to extract colors.

        Returns
        -------
        set[str]
            A set of hex color strings representing all unique colors present in the pixmap.
        """
        arr = Pixmap.to_numpy(pixmap)
        argb32_set = np.unique_values(arr)
        color_names = {QtGui.QColor.fromRgba(x).name() for x in argb32_set}
        return color_names

    @staticmethod
    def color_names_exclude_defaults(pixmap: QtGui.QPixmap) -> set[str]:
        """
        Return a set of all unique colors in a QPixmap, excluding default colors.

        This function extracts all unique colors from the pixmap and removes
        predefined default colors such as background, cursor, and range colors.

        Parameters
        ----------
        pixmap : QtGui.QPixmap
            The QPixmap from which to extract colors.

        Returns
        -------
        set[str]
            A set of hex color strings representing all unique colors in the pixmap,
            excluding the default background, cursor, and range colors.
        """
        defaults = {Pixmap.COLOR_BACKGROUND, Pixmap.COLOR_CURSOR, Pixmap.COLOR_RANGE}
        return Pixmap.color_names(pixmap) - defaults

    @staticmethod
    def color_map(pixmap: QtGui.QPixmap) -> dict[int, list[str]]:
        """
        Convert a QPixmap into a per-line mapping of pixel colors as hex strings.

        This function efficiently maps each pixel in the pixmap to its corresponding
        hex color string using NumPy. It avoids Python loops over individual pixels
        by vectorizing the conversion from ARGB32 values to color strings.

        Parameters
        ----------
        pixmap : QtGui.QPixmap
            The QPixmap to convert.

        Returns
        -------
        dict[int, list[str]]
            A dictionary where:
            - Keys are row indices (0-based).
            - Values are lists of hex color strings representing the colors of
            pixels in that row, ordered by column.
        """
        arr = Pixmap.to_numpy(pixmap)  # shape (H, W), dtype uint32

        # 1. Map all unique ARGB values to hex strings
        unique_vals = np.unique(arr)  # returns sortes values
        lookup = np.array([QtGui.QColor.fromRgba(int(val)).name() for val in unique_vals])

        # 2. Create an index array
        # np.searchsorted requires sorted unique_vals
        indices = np.searchsorted(unique_vals, arr)

        # 3. Map indices to hex strings
        hex_array = lookup[indices]  # shape (H, W) (hex strings)

        return {y: list(row) for y, row in enumerate(hex_array)}

    @staticmethod
    def cursors_x(pixmap: QtGui.QPixmap, threshold: float = 0.3) -> list[int]:
        """
        Find the x-coordinates of vertical cursors in a QPixmap, including dashed cursors.

        A column is considered a vertical cursor if the fraction of pixels matching
        the cursor color exceeds the given threshold.

        Parameters
        ----------
        pixmap : QtGui.QPixmap
            The pixmap to scan for vertical cursor pixels.
        threshold : float
            Minimum fraction of pixels in a column that must match the cursor color
            to consider it a vertical cursor (between 0 and 1).

        Returns
        -------
        list[int]
            List of x-coordinates corresponding to detected vertical cursors.
        """
        arr = Pixmap.to_numpy(pixmap)
        cursor_color = QtGui.QColor(Pixmap.COLOR_CURSOR).rgba()

        # Count pixels matching the cursor color per column
        match_counts = np.sum(arr == cursor_color, axis=0)

        # Determine fraction of pixels in each column that match
        fraction = match_counts / arr.shape[0]

        # Columns exceeding the threshold are considered vertical cursors
        cols_with_cursor = np.nonzero(fraction >= threshold)[0]

        return [int(x) for x in cols_with_cursor]

    @staticmethod
    def search_signal_extremes_by_ax(
        pixmap: QtGui.QPixmap, signal_color: str | QtGui.QColor, ax: Literal["X", "Y"]
    ) -> tuple[int, int] | None:
        """
        Find the start and end indices of a colored signal in a QPixmap along a specified axis.

        This function returns the first and last row or column where the signal color appears,
        using NumPy for fast pixel scanning.

        Parameters
        ----------
        pixmap : QtGui.QPixmap
            The pixmap containing the signal.
        signal_color : str or QtGui.QColor
            The color of the signal to search for, as a hex string or QColor.
        ax : {"X", "Y"}
            The axis along which to find extremes:
            - "X": returns [start_column, end_column]
            - "Y": returns [start_row, end_row]

        Returns
        -------
        tuple[int, int] | None
            Two-element tuple with the start and end index along the chosen axis, or None if color not found.
        """
        # Normalize signal color to hex string
        if isinstance(signal_color, str):
            signal_color = QtGui.QColor(signal_color)

        arr = Pixmap.to_numpy(pixmap)  # shape (H, W), dtype uint32
        # Convert signal color to ARGB uint32
        signal_val = QtGui.QColor(signal_color).rgba()

        if ax.upper() == "X":
            mask = np.any(arr == signal_val, axis=0)  # columns containing signal
            indices = np.flatnonzero(mask)
        elif ax.upper() == "Y":
            mask = np.any(arr == signal_val, axis=1)  # rows containing signal
            indices = np.flatnonzero(mask)
        else:
            raise ValueError("ax must be 'X' or 'Y'")

        if len(indices) == 0:
            return None

        return (int(indices[0]), int(indices[-1]))


class OpenFileContextManager:
    """
    Generic class for opening a file using context manager.
    Methods:
        __enter__: return opened file object.
        __exit__: close file object. If exc_type, exc_val, exc_tb, raise exception.
    """

    def __init__(self, file_path: str | pathlib.Path):
        """
        Parameters
        ----------
        file_path: file path as str or pathlib.Path object
        """
        self.file = None
        if isinstance(file_path, str):
            self._file_path = pathlib.Path(file_path)
        else:
            self._file_path = file_path

        assert self._file_path.exists(), "Provided file does not exist"

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.file.close()
        for exc in (exc_type, exc_val, exc_tb):
            if exc is not None:
                print(exc)


class OpenMDF(OpenFileContextManager):
    """
    Open MDF file using context manager.
    """

    def __enter__(self):
        self.file = mdf.MDF(self._file_path, process_bus_logging=("process_bus_logging", True))
        return self.file


class OpenHDF5(OpenFileContextManager):
    """
    Open HDF5 file using context manager.
    """

    def __enter__(self):
        self.file = HDF5(self._file_path)
        return self.file


class DBC:
    class BO:
        def __init__(self, lines: Iterable[str]):
            for line in lines:
                if line.startswith("BO_ "):
                    self.data = line

        def __repr__(self):
            return self.data

        @property
        def name(self) -> str:
            if self.data:
                return self.data.split()[2].strip(":")

        @property
        def id(self) -> str:
            if self.data:
                return self.data.split()[1]

        @property
        def data_length(self) -> int:
            if self.data:
                return int(self.data.split()[3])

    class SG:
        def __init__(self, name: str, lines: Iterable[str]):
            self.name = name
            self.data = None
            for line in lines:
                if "SG_ " in line and name in line:
                    self.data = line.split("SG_ ")[1]

        def __repr__(self):
            return self.data

        @property
        def unit(self) -> str:
            return self.data.split('"')[1]  # middle value, ex: "blah blah "unit" blah

        @property
        def bit_count(self) -> int:
            return int(self.data.split("|")[1].split("@")[0])  # is complicated this CAN :)

        @property
        def conversion_a(self) -> float:
            return float(self.data.split("(")[1].split(")")[0].split(",")[0])

        @property
        def conversion_b(self) -> float:
            return float(self.data.split("(")[1].split(")")[0].split(",")[1])



"""asammdf utility functions and classes"""

from collections.abc import Callable, Iterator
from functools import lru_cache
import logging
import mmap
import multiprocessing
from pathlib import Path
from random import randint
import re
import string
from struct import Struct
import subprocess
import sys
from tempfile import TemporaryDirectory
from time import perf_counter
from types import TracebackType
import typing
from typing import Final, Literal, Optional, Union
import xml.etree.ElementTree as ET

from canmatrix.canmatrix import CanMatrix, matrix_class
import canmatrix.formats
import numpy as np
from numpy import arange, bool_, interp, where
from numpy.typing import NDArray
import pandas as pd
from pandas import Series
from typing_extensions import (
    Any,
    Buffer,
    LiteralString,
    overload,
    ParamSpec,
    Protocol,
    runtime_checkable,
    TypedDict,
    TypeIs,
    TypeVar,
    Unpack,
)

from . import v2_v3_constants as v3c
from . import v4_constants as v4c
from .blocks_common import UnpackFrom
from .options import GLOBAL_OPTIONS
from .types import StrPath

try:
    from cchardet import detect
except:
    try:
        from chardet import detect
    except:

        class DetectDict(TypedDict):
            encoding: str | None

        def detect(text: bytes) -> DetectDict:
            encoding: str | None
            for encoding in ("utf-8", "latin-1", "cp1250", "cp1252"):
                try:
                    text.decode(encoding)
                    break
                except:
                    continue
            else:
                encoding = None
            return {"encoding": encoding}


class Terminated(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__("terminated by user", *args)


THREAD_COUNT: Final = max(multiprocessing.cpu_count() - 1, 1)
target_byte_order: Final = "<=" if sys.byteorder == "little" else ">="


UINT8_u: Callable[[Buffer], tuple[int]] = Struct("<B").unpack
UINT16_u: Callable[[Buffer], tuple[int]] = Struct("<H").unpack
UINT32_p = Struct("<I").pack
UINT32_u: Callable[[Buffer], tuple[int]] = Struct("<I").unpack
UINT64_u: Callable[[Buffer], tuple[int]] = Struct("<Q").unpack
UINT8_uf: UnpackFrom[tuple[int]] = Struct("<B").unpack_from
UINT16_uf: UnpackFrom[tuple[int]] = Struct("<H").unpack_from
UINT32_uf: UnpackFrom[tuple[int]] = Struct("<I").unpack_from
UINT64_uf: UnpackFrom[tuple[int]] = Struct("<Q").unpack_from
FLOAT64_u: Callable[[Buffer], tuple[float]] = Struct("<d").unpack
FLOAT64_uf: UnpackFrom[tuple[float]] = Struct("<d").unpack_from
TWO_UINT64_u: Callable[[Buffer], tuple[int, int]] = Struct("<2Q").unpack
TWO_UINT64_uf: UnpackFrom[tuple[int, int]] = Struct("<2Q").unpack_from
BLK_COMMON_uf: UnpackFrom[tuple[bytes, int]] = Struct("<4s4xQ").unpack_from
BLK_COMMON_u: Callable[[Buffer], tuple[bytes, int]] = Struct("<4s4xQ8x").unpack

EMPTY_TUPLE: Final = ()

_xmlns_pattern = re.compile(' xmlns="[^"]*"')

logger = logging.getLogger("asammdf")

__all__ = [
    "CHANNEL_COUNT",
    "CONVERT",
    "MDF2_VERSIONS",
    "MDF3_VERSIONS",
    "MDF4_VERSIONS",
    "MERGE",
    "SUPPORTED_VERSIONS",
    "ChannelsDB",
    "MdfException",
    "UniqueDB",
    "extract_xml_comment",
    "fmt_to_datatype_v3",
    "fmt_to_datatype_v4",
    "get_fmt_v3",
    "get_fmt_v4",
    "get_text_v4",
    "matlab_compatible",
    "validate_version_argument",
]

_channel_count = (1000, 2000, 10000, 20000)
CHANNEL_COUNT: Final = arange(0, 20000, 1000, dtype=np.dtype("<u4"))

_convert = (10 * 2**20, 20 * 2**20, 30 * 2**20, 40 * 2**20)
CONVERT: Final = interp(CHANNEL_COUNT, _channel_count, _convert).astype("<u4")

_merge = (10 * 2**20, 20 * 2**20, 35 * 2**20, 60 * 2**20)
MERGE: Final = interp(CHANNEL_COUNT, _channel_count, _merge).astype("<u4")

MDF2_VERSIONS: Final[tuple[LiteralString, ...]] = ("2.00", "2.10", "2.14")
MDF3_VERSIONS: Final[tuple[LiteralString, ...]] = ("3.00", "3.10", "3.20", "3.30")
MDF4_VERSIONS: Final[tuple[LiteralString, ...]] = ("4.00", "4.10", "4.11", "4.20")
SUPPORTED_VERSIONS: Final = MDF2_VERSIONS + MDF3_VERSIONS + MDF4_VERSIONS


ALLOWED_MATLAB_CHARS: Final = set(string.ascii_letters + string.digits + "_")


class MdfException(Exception):
    """MDF Exception class."""

    def __repr__(self) -> str:
        return f"asammdf MdfException: {self.args[0]}"


def extract_xml_comment(comment: str) -> str:
    """Extract *TX* tag or otherwise the *common_properties* from an XML comment.

    Parameters
    ----------
    comment : str
        XML string comment.

    Returns
    -------
    comment : str
        Extracted string.
    """

    comment = comment.replace(' xmlns="http://www.asam.net/mdf/v4"', "")
    try:
        comment_elem = ET.fromstring(comment)
        match = comment_elem.find(".//TX")
        if match is None:
            common_properties = comment_elem.find(".//common_properties")
            if common_properties is not None:
                comments: list[str] = []
                for e in common_properties:
                    field = f"{e.get('name')}: {e.text}"
                    comments.append(field)
                comment = "\n".join(comments)
            else:
                comment = ""
        else:
            comment = match.text or ""
    except ET.ParseError:
        pass

    return comment


def matlab_compatible(name: str) -> str:
    """Make a channel name compatible with Matlab variable naming.

    Parameters
    ----------
    name : str
        Channel name.

    Returns
    -------
    compatible_name : str
        Channel name compatible with Matlab.
    """

    compatible_names = [ch if ch in ALLOWED_MATLAB_CHARS else "_" for ch in name]
    compatible_name = "".join(compatible_names)

    if compatible_name[0] not in string.ascii_letters:
        compatible_name = "M_" + compatible_name

    # max variable name is 63 and 3 chars are reserved
    # for get_unique_name in case of multiple channel name occurrence
    return compatible_name[:60]


@runtime_checkable
class FileLike(Protocol):
    def __iter__(self) -> Iterator[bytes]: ...
    def close(self) -> None: ...
    def read(self, size: int | None = -1, /) -> bytes: ...
    def seek(self, target: int, whence: int = 0, /) -> int: ...
    def tell(self) -> int: ...
    def write(self, buffer: Buffer, /) -> int: ...


class BlockKwargs(TypedDict, total=False):
    stream: FileLike | mmap.mmap
    mapped: bool
    address: int


@overload
def get_text_v3(
    address: int,
    stream: FileLike | mmap.mmap,
    mapped: bool = ...,
    decode: Literal[True] = ...,
) -> str: ...


@overload
def get_text_v3(
    address: int,
    stream: FileLike | mmap.mmap,
    mapped: bool = ...,
    *,
    decode: Literal[False],
) -> bytes: ...


@overload
def get_text_v3(address: int, stream: FileLike | mmap.mmap, mapped: bool = ..., decode: bool = ...) -> bytes | str: ...


def get_text_v3(address: int, stream: FileLike | mmap.mmap, mapped: bool = False, decode: bool = True) -> bytes | str:
    """Faster way to extract strings from MDF versions 2 and 3 TextBlock.

    Parameters
    ----------
    address : int
        TextBlock address.
    stream : handle
        File IO handle.
    mapped : bool, default False
        Flag for mapped stream.
    decode : bool, default True
        Option to auto-detect character encoding and return decoded str instead
        of raw bytes.

    Returns
    -------
    text : str | bytes
        Unicode string or bytes object depending on the `decode` argument.
    """

    if address == 0:
        return "" if decode else b""

    if mapped:
        block_id = stream[address : address + 2]
        if block_id != b"TX":
            return "" if decode else b""
        (size,) = UINT16_uf(stream, address + 2)
        text_bytes = stream[address + 4 : address + size].split(b"\0", 1)[0].strip(b" \r\t\n")
    else:
        stream.seek(address)
        block_id = stream.read(2)
        if block_id != b"TX":
            return "" if decode else b""
        size = UINT16_u(stream.read(2))[0] - 4
        text_bytes = stream.read(size).split(b"\0", 1)[0].strip(b" \r\t\n")

    text: bytes | str

    if decode:
        try:
            text = text_bytes.decode("latin-1")
        except UnicodeDecodeError:
            encoding = detect(text_bytes)["encoding"]
            if encoding:
                try:
                    text = text_bytes.decode(encoding, "ignore")
                except:
                    text = "<!text_decode_error>"
            else:
                text = "<!text_decode_error>"
    else:
        text = text_bytes

    return text


@overload
def get_text_v4(
    address: int,
    stream: FileLike | mmap.mmap,
    mapped: bool = ...,
    decode: Literal[True] = ...,
    *,
    file_limit: int | float = float("inf"),
) -> str: ...


@overload
def get_text_v4(
    address: int,
    stream: FileLike | mmap.mmap,
    mapped: bool = ...,
    *,
    decode: Literal[False],
    file_limit: int | float = float("inf"),
) -> bytes: ...


@overload
def get_text_v4(
    address: int,
    stream: FileLike | mmap.mmap,
    mapped: bool = ...,
    decode: bool = ...,
    *,
    file_limit: int | float = float("inf"),
) -> bytes | str: ...


def get_text_v4(
    address: int,
    stream: FileLike | mmap.mmap,
    mapped: bool = False,
    decode: bool = True,
    *,
    file_limit: int | float = float("inf"),
) -> bytes | str:
    """Faster way to extract strings from MDF version 4 TextBlock.

    Parameters
    ----------
    address : int
        TextBlock address.
    stream : handle
        File IO handle.
    mapped : bool, default False
        Flag for mapped stream.
    decode : bool, default True
        Option to auto-detect character encoding and return decoded str instead
        of raw bytes.

    Returns
    -------
    text : str | bytes
        Unicode string or bytes object depending on the `decode` argument.
    """



    if address == 0:
        return "" if decode else b""

    if mapped:
        if address + 16 > file_limit:
            handle_incomplete_block(address)
            return "" if decode else b""

        block_id, size = BLK_COMMON_uf(stream, address)
        if block_id not in (b"##TX", b"##MD"):
            return "" if decode else b""

        if (end_of_string := address + size) > file_limit:
            handle_incomplete_block(address)
            return "" if decode else b""

        text_bytes = stream[address + 24 : end_of_string].split(b"\0", 1)[0].strip(b" \r\t\n")
    else:
        if address + 24 > file_limit:
            handle_incomplete_block(address)
            return "" if decode else b""
        stream.seek(address)
        block_id, size = BLK_COMMON_u(stream.read(24))
        if block_id not in (b"##TX", b"##MD"):
            return "" if decode else b""

        if address + size > file_limit:
            handle_incomplete_block(address)
            return "" if decode else b""

        text_bytes = stream.read(size - 24).split(b"\0", 1)[0].strip(b" \r\t\n")

    if not decode:
        return text_bytes

    try:
        decoded_text = text_bytes.decode("utf-8")
    except UnicodeDecodeError:
        encoding = detect(text_bytes)["encoding"]
        if encoding:
            try:
                decoded_text = text_bytes.decode(encoding, "ignore")
            except:
                decoded_text = "<!text_decode_error>"
        else:
            decoded_text = "<!text_decode_error>"

    return decoded_text


def sanitize_xml(text: str) -> str:
    return re.sub(_xmlns_pattern, "", text)


def extract_display_names(comment: str) -> dict[str, str]:
    comment = comment.replace(' xmlns="http://www.asam.net/mdf/v4"', "")
    display_names = {}
    if comment.startswith("<CN") and "<names>" in comment:
        try:
            start = comment.index("<names>")
            end = comment.index("</names>") + 8
            names = ET.fromstring(comment[start:end])
            for i, elem in enumerate(names.iter()):
                if i == 0:
                    continue
                if elem.text is None:
                    raise ValueError("text is None")
                display_names[elem.text.strip(" \t\r\n\v\0")] = elem.tag

        except:
            pass

    return display_names


class EncryptionInfo(TypedDict, total=False):
    encrypted: bool
    algorithm: str
    original_md5_sum: str
    original_size: int


def extract_encryption_information(comment: str) -> EncryptionInfo:
    info: EncryptionInfo = {}
    comment = comment.replace(' xmlns="http://www.asam.net/mdf/v4"', "")
    if comment.startswith("<ATcomment") and "<encrypted>" in comment:
        try:
            comment_elem = ET.fromstring(comment)
            for match in comment_elem.findall(".//extensions/extension"):
                elem = match.find("encrypted")
                if elem is None:
                    raise RuntimeError("cannot find 'encrypted' Element")
                if elem.text is None:
                    raise RuntimeError("text is None")
                encrypted = elem.text.strip().lower() == "true"
                elem = match.find("algorithm")
                if elem is None:
                    raise RuntimeError("cannot find 'algorithm' Element")
                if elem.text is None:
                    raise RuntimeError("text is None")
                algorithm = elem.text.strip().lower()
                elem = match.find("original_md5_sum")
                if elem is None:
                    raise RuntimeError("cannot find 'original_md5_sum' Element")
                if elem.text is None:
                    raise RuntimeError("text is None")
                original_md5_sum = elem.text.strip().lower()
                elem = match.find("original_size")
                if elem is None:
                    raise RuntimeError("cannot find 'original_size' Element")
                if elem.text is None:
                    raise RuntimeError("text is None")
                original_size = int(elem.text)

                info["encrypted"] = encrypted
                info["algorithm"] = algorithm
                info["original_md5_sum"] = original_md5_sum
                info["original_size"] = original_size
                break
        except:
            pass

    return info


def extract_ev_tool(comment: str) -> str:
    tool = ""
    comment = comment.replace(' xmlns="http://www.asam.net/mdf/v4"', "")
    try:
        comment_elem = ET.fromstring(comment)
        match = comment_elem.find(".//tool")
        if match is None:
            tool = ""
        else:
            tool = match.text or ""
    except:
        pass

    return tool


@lru_cache(maxsize=1024)
def get_fmt_v3(data_type: int, size: int, byte_order: int = v3c.BYTE_ORDER_INTEL) -> str:
    """Convert MDF versions 2 and 3 channel data type to numpy dtype format
    string.

    Parameters
    ----------
    data_type : int
        MDF channel data type.
    size : int
        Data bit size.
    byte_order : int, default 0 (BYTE_ORDER_INTEL)
        Integer code for byte order (endianness).

    Returns
    -------
    fmt : str
        Numpy compatible data type format string.
    """
    if data_type in (v3c.DATA_TYPE_STRING, v3c.DATA_TYPE_BYTEARRAY):
        size = size // 8
        if data_type == v3c.DATA_TYPE_STRING:
            fmt = f"S{size}"
        else:
            fmt = f"({size},)u1"
    else:
        if size > 64 and data_type in (
            v3c.DATA_TYPE_UNSIGNED_INTEL,
            v3c.DATA_TYPE_UNSIGNED,
            v3c.DATA_TYPE_UNSIGNED_MOTOROLA,
            v3c.DATA_TYPE_SIGNED_INTEL,
            v3c.DATA_TYPE_SIGNED,
            v3c.DATA_TYPE_SIGNED_MOTOROLA,
        ):
            fmt = f"({size // 8},)u1"
        else:
            if size <= 8:
                size = 1
            elif size <= 16:
                size = 2
            elif size <= 32:
                size = 4
            elif size <= 64:
                size = 8
            else:
                size = size // 8

            match data_type:
                case v3c.DATA_TYPE_UNSIGNED_INTEL:
                    fmt = f"<u{size}"

                case v3c.DATA_TYPE_UNSIGNED:
                    if byte_order == v3c.BYTE_ORDER_INTEL:
                        fmt = f"<u{size}"
                    else:
                        fmt = f">u{size}"

                case v3c.DATA_TYPE_UNSIGNED_MOTOROLA:
                    fmt = f">u{size}"

                case v3c.DATA_TYPE_SIGNED_INTEL:
                    fmt = f"<i{size}"

                case v3c.DATA_TYPE_SIGNED:
                    if byte_order == v3c.BYTE_ORDER_INTEL:
                        fmt = f"<i{size}"
                    else:
                        fmt = f">i{size}"

                case v3c.DATA_TYPE_SIGNED_MOTOROLA:
                    fmt = f">i{size}"

                case v3c.DATA_TYPE_FLOAT_INTEL | v3c.DATA_TYPE_DOUBLE_INTEL:
                    fmt = f"<f{size}"

                case v3c.DATA_TYPE_FLOAT_MOTOROLA | v3c.DATA_TYPE_DOUBLE_MOTOROLA:
                    fmt = f">f{size}"

                case v3c.DATA_TYPE_FLOAT | v3c.DATA_TYPE_DOUBLE:
                    if byte_order == v3c.BYTE_ORDER_INTEL:
                        fmt = f"<f{size}"
                    else:
                        fmt = f">f{size}"

    return fmt


@lru_cache(maxsize=1024)
def get_fmt_v4(data_type: int, size: int, channel_type: int = v4c.CHANNEL_TYPE_VALUE) -> str:
    """Convert MDF version 4 channel data type to numpy dtype format string.

    Parameters
    ----------
    data_type : int
        MDF channel data type.
    size : int
        Data bit size.
    channel_type : int, default 0 (CHANNEL_TYPE_VALUE)
        MDF channel type.

    Returns
    -------
    fmt : str
        Numpy compatible data type format string.
    """
    if data_type in v4c.NON_SCALAR_TYPES:
        size = size // 8 or 1

        if data_type in (
            v4c.DATA_TYPE_BYTEARRAY,
            v4c.DATA_TYPE_MIME_STREAM,
            v4c.DATA_TYPE_MIME_SAMPLE,
        ):
            if channel_type == v4c.CHANNEL_TYPE_VALUE:
                fmt = f"({size},)u1"
            else:
                fmt = f"<u{size}"

        elif data_type in v4c.STRING_TYPES:
            if channel_type == v4c.CHANNEL_TYPE_VALUE:
                fmt = f"S{size}"
            else:
                fmt = f"<u{size}"

        elif data_type == v4c.DATA_TYPE_CANOPEN_DATE:
            fmt = "V7"

        elif data_type == v4c.DATA_TYPE_CANOPEN_TIME:
            fmt = "V6"

    elif channel_type in v4c.VIRTUAL_TYPES:
        match data_type:
            case v4c.DATA_TYPE_UNSIGNED_INTEL:
                fmt = "<u8"

            case v4c.DATA_TYPE_UNSIGNED_MOTOROLA:
                fmt = ">u8"

            case v4c.DATA_TYPE_SIGNED_INTEL:
                fmt = "<i8"

            case v4c.DATA_TYPE_SIGNED_MOTOROLA:
                fmt = ">i8"

            case v4c.DATA_TYPE_REAL_INTEL:
                fmt = "<f8"

            case v4c.DATA_TYPE_REAL_MOTOROLA:
                fmt = ">f8"
            case v4c.DATA_TYPE_COMPLEX_INTEL:
                fmt = "<c8"
            case v4c.DATA_TYPE_COMPLEX_MOTOROLA:
                fmt = ">c8"

    else:
        if size > 64 and data_type in (
            v4c.DATA_TYPE_UNSIGNED_INTEL,
            v4c.DATA_TYPE_UNSIGNED_MOTOROLA,
            v4c.DATA_TYPE_SIGNED_INTEL,
            v4c.DATA_TYPE_SIGNED_MOTOROLA,
        ):
            fmt = f"({size // 8},)u1"
        else:
            if size <= 8:
                size = 1
            elif size <= 16:
                size = 2
            elif size <= 32:
                size = 4
            elif size <= 64:
                size = 8
            else:
                size = size // 8

            match data_type:
                case v4c.DATA_TYPE_UNSIGNED_INTEL:
                    fmt = f"<u{size}"

                case v4c.DATA_TYPE_UNSIGNED_MOTOROLA:
                    fmt = f">u{size}"

                case v4c.DATA_TYPE_SIGNED_INTEL:
                    fmt = f"<i{size}"

                case v4c.DATA_TYPE_SIGNED_MOTOROLA:
                    fmt = f">i{size}"

                case v4c.DATA_TYPE_REAL_INTEL:
                    fmt = f"<f{size}"

                case v4c.DATA_TYPE_REAL_MOTOROLA:
                    fmt = f">f{size}"
                case v4c.DATA_TYPE_COMPLEX_INTEL:
                    fmt = f"<c{size}"
                case v4c.DATA_TYPE_COMPLEX_MOTOROLA:
                    fmt = f">c{size}"

    return fmt


@lru_cache(maxsize=1024)
def fmt_to_datatype_v3(fmt: np.dtype[Any], shape: tuple[int, ...], array: bool = False) -> tuple[int, int]:
    """Convert numpy dtype format string to MDF versions 2 and 3
    channel data type and size.

    Parameters
    ----------
    fmt : numpy.dtype
        Numpy data type.
    shape : tuple
        Numpy array shape.
    array : bool, default False
        Disambiguate between bytearray and channel array.

    Returns
    -------
    data_type, size : int, int
        Integer data type as defined by ASAM MDF and bit size.
    """
    byteorder = fmt.byteorder
    if byteorder in "=|":
        byteorder = "<" if sys.byteorder == "little" else ">"
    size = fmt.itemsize * 8
    kind = fmt.kind

    if not array and shape[1:] and fmt.itemsize == 1 and kind == "u":
        data_type = v3c.DATA_TYPE_BYTEARRAY
        for dim in shape[1:]:
            size *= dim
    else:
        match kind:
            case "u":
                if byteorder == "<":
                    data_type = v3c.DATA_TYPE_UNSIGNED_INTEL
                else:
                    data_type = v3c.DATA_TYPE_UNSIGNED_MOTOROLA
            case "i":
                if byteorder == "<":
                    data_type = v3c.DATA_TYPE_SIGNED_INTEL
                else:
                    data_type = v3c.DATA_TYPE_SIGNED_MOTOROLA
            case "f":
                if byteorder == "<":
                    if size == 32:
                        data_type = v3c.DATA_TYPE_FLOAT
                    else:
                        data_type = v3c.DATA_TYPE_DOUBLE
                else:
                    if size == 32:
                        data_type = v3c.DATA_TYPE_FLOAT_MOTOROLA
                    else:
                        data_type = v3c.DATA_TYPE_DOUBLE_MOTOROLA
            case "S" | "V":
                data_type = v3c.DATA_TYPE_STRING
            case "b":
                data_type = v3c.DATA_TYPE_UNSIGNED_INTEL
                size = 1
            case _:
                message = f"Unknown type: dtype={fmt}, shape={shape}"
                logger.exception(message)
                raise MdfException(message)

    return data_type, size


@lru_cache(maxsize=1024)
def info_to_datatype_v4(signed: bool, little_endian: bool) -> int:
    """Map CAN signal to MDF integer types.

    Parameters
    ----------
    signed : bool
        Signal is flagged as signed in the CAN database.
    little_endian : bool
        Signal is flagged as little-endian (Intel) in the CAN database.

    Returns
    -------
    datatype : int
        Integer code for MDF channel data type.
    """

    if signed:
        if little_endian:
            datatype = v4c.DATA_TYPE_SIGNED_INTEL
        else:
            datatype = v4c.DATA_TYPE_SIGNED_MOTOROLA
    else:
        if little_endian:
            datatype = v4c.DATA_TYPE_UNSIGNED_INTEL
        else:
            datatype = v4c.DATA_TYPE_UNSIGNED_MOTOROLA

    return datatype


@lru_cache(maxsize=1024)
def fmt_to_datatype_v4(fmt: np.dtype[Any], shape: tuple[int, ...], array: bool = False) -> tuple[int, int]:
    """Convert numpy dtype format string to MDF version 4 channel data
    type and size.

    Parameters
    ----------
    fmt : numpy.dtype
        Numpy data type.
    shape : tuple
        Numpy array shape.
    array : bool, default False
        Disambiguate between bytearray and channel array.

    Returns
    -------
    data_type, size : int, int
        Integer data type as defined by ASAM MDF and bit size.
    """
    byteorder = fmt.byteorder
    if byteorder in "=|":
        byteorder = "<" if sys.byteorder == "little" else ">"
    size = fmt.itemsize * 8
    kind = fmt.kind

    if not array and len(shape) > 1 and size == 8 and kind == "u":
        data_type = v4c.DATA_TYPE_BYTEARRAY
        for dim in shape[1:]:
            size *= dim

    else:
        match kind:
            case "u":
                if byteorder == "<":
                    data_type = v4c.DATA_TYPE_UNSIGNED_INTEL
                else:
                    data_type = v4c.DATA_TYPE_UNSIGNED_MOTOROLA
            case "i":
                if byteorder == "<":
                    data_type = v4c.DATA_TYPE_SIGNED_INTEL
                else:
                    data_type = v4c.DATA_TYPE_SIGNED_MOTOROLA
            case "f":
                if byteorder == "<":
                    data_type = v4c.DATA_TYPE_REAL_INTEL
                else:
                    data_type = v4c.DATA_TYPE_REAL_MOTOROLA
            case "S" | "V":
                data_type = v4c.DATA_TYPE_STRING_LATIN_1
            case "b":
                data_type = v4c.DATA_TYPE_UNSIGNED_INTEL
                size = 1
            case "c":
                if byteorder == "<":
                    data_type = v4c.DATA_TYPE_COMPLEX_INTEL
                else:
                    data_type = v4c.DATA_TYPE_COMPLEX_MOTOROLA
            case _:
                message = f"Unknown type: dtype={fmt}, shape={shape}"
                logger.exception(message)
                raise MdfException(message)

    return data_type, size


def as_non_byte_sized_signed_int(integer_array: NDArray[Any], bit_length: int) -> NDArray[Any]:
    """The MDF spec allows values to be encoded as integers that aren't
    byte-sized. Numpy only knows how to do two's complement on byte-sized
    integers (i.e. int16, int32, int64, etc.), so we have to calculate two's
    complement ourselves in order to handle signed integers with unconventional
    lengths.

    Parameters
    ----------
    integer_array : np.ndarray
        Array of integers to apply two's complement to.
    bit_length : int
        Number of bits to sample from the array.

    Returns
    -------
    integer_array : np.ndarray
        Signed integer array with non-byte-sized two's complement applied.
    """

    if integer_array.flags.writeable:
        integer_array &= (1 << bit_length) - 1  # Zero out the unwanted bits
        truncated_integers = integer_array
    else:
        truncated_integers = integer_array & ((1 << bit_length) - 1)  # Zero out the unwanted bits
    return where(
        truncated_integers >> bit_length - 1,  # sign bit as a truth series (True when negative)
        (2**bit_length - truncated_integers) * np.int8(-1),  # when negative, do two's complement
        truncated_integers,  # when positive, return the truncated int
    )


def count_channel_groups(
    stream: FileLike | mmap.mmap, include_channels: bool = False, mapped: bool = False
) -> tuple[int, int]:
    """Count all channel groups as fast as possible. This is used to provide
    reliable progress information when loading a file using the GUI.

    Parameters
    ----------
    stream : file handle
        Opened file handle.
    include_channels : bool, default False
        Also count channels.
    mapped : bool, default False
        Flag for mapped stream.

    Returns
    -------
    count : int
        Channel group count.
    """

    count = 0
    ch_count = 0

    stream.seek(0, 2)
    file_limit = stream.tell()

    stream.seek(64)
    blk_id = stream.read(2)
    if blk_id == b"HD":
        version = 3
    else:
        blk_id += stream.read(2)
        if blk_id == b"##HD":
            version = 4
        else:
            raise MdfException(f"'{getattr(stream, 'name', stream)}' is not a valid MDF file")

    if version >= 4:
        if mapped:
            dg_addr = UINT64_uf(stream, 88)[0]
            while dg_addr:
                stream.seek(dg_addr + 32)
                cg_addr = UINT64_uf(stream, dg_addr + 32)[0]
                while cg_addr:
                    count += 1
                    if include_channels:
                        ch_addr = UINT64_uf(stream, cg_addr + 32)[0]
                        while ch_addr:
                            ch_count += 1
                            ch_addr = UINT64_uf(stream, ch_addr + 24)[0]
                            if ch_addr >= file_limit:
                                raise MdfException("File is a corrupted MDF file - Invalid CH block address")

                    cg_addr = UINT64_uf(stream, cg_addr + 24)[0]
                    if cg_addr >= file_limit:
                        raise MdfException("File is a corrupted MDF file - Invalid CG block address")

                dg_addr = UINT64_uf(stream, dg_addr + 24)[0]
                if dg_addr >= file_limit:
                    raise MdfException("File is a corrupted MDF file - Invalid DG block address")
        else:
            stream.seek(88, 0)
            dg_addr = UINT64_u(stream.read(8))[0]
            while dg_addr:
                stream.seek(dg_addr + 32)
                cg_addr = UINT64_u(stream.read(8))[0]
                while cg_addr:
                    count += 1
                    if include_channels:
                        stream.seek(cg_addr + 32)
                        ch_addr = UINT64_u(stream.read(8))[0]
                        while ch_addr:
                            ch_count += 1
                            stream.seek(ch_addr + 24)
                            ch_addr = UINT64_u(stream.read(8))[0]
                            if ch_addr >= file_limit:
                                raise MdfException("File is a corrupted MDF file - Invalid CH block address")

                    stream.seek(cg_addr + 24)
                    cg_addr = UINT64_u(stream.read(8))[0]
                    if cg_addr >= file_limit:
                        raise MdfException("File is a corrupted MDF file - Invalid CG block address")

                stream.seek(dg_addr + 24)
                dg_addr = UINT64_u(stream.read(8))[0]

                if dg_addr >= file_limit:
                    raise MdfException("File is a corrupted MDF file - Invalid DG block address")

    else:
        stream.seek(68, 0)
        dg_addr = UINT32_u(stream.read(4))[0]
        while dg_addr:
            stream.seek(dg_addr + 8)
            cg_addr = UINT32_u(stream.read(4))[0]
            while cg_addr:
                count += 1
                if include_channels:
                    stream.seek(cg_addr + 8)
                    ch_addr = UINT32_u(stream.read(4))[0]
                    while ch_addr:
                        ch_count += 1
                        stream.seek(ch_addr + 4)
                        ch_addr = UINT32_u(stream.read(4))[0]
                        if ch_addr >= file_limit:
                            raise MdfException("File is a corrupted MDF file - Invalid CH block address")

                stream.seek(cg_addr + 4)
                cg_addr = UINT32_u(stream.read(4))[0]
                if cg_addr >= file_limit:
                    raise MdfException("File is a corrupted MDF file - Invalid CG block address")

            stream.seek(dg_addr + 4)
            dg_addr = UINT32_u(stream.read(4))[0]

            if dg_addr >= file_limit:
                raise MdfException("File is a corrupted MDF file - Invalid DG block address")

    return count, ch_count


@overload
def validate_version_argument(version: v3c.Version2, hint: Literal[2]) -> v3c.Version2: ...


@overload
def validate_version_argument(version: v3c.Version2 | v3c.Version, hint: Literal[3]) -> v3c.Version2 | v3c.Version: ...


@overload
def validate_version_argument(version: v4c.Version, hint: Literal[4] = ...) -> v4c.Version: ...


@overload
def validate_version_argument(version: str, hint: int = ...) -> v3c.Version2 | v3c.Version | v4c.Version: ...


def validate_version_argument(version: str, hint: int = 4) -> v3c.Version2 | v3c.Version | v4c.Version:
    """Validate the version argument against the supported MDF versions. The
    default version used depends on the hint MDF major revision.

    Parameters
    ----------
    version : str
        Requested MDF version.
    hint : int, default 4
        MDF revision hint.

    Returns
    -------
    valid_version : str
        Valid version.
    """
    valid_version: v3c.Version2 | v3c.Version | v4c.Version
    if version not in SUPPORTED_VERSIONS:
        if hint == 2:
            valid_version = "2.14"
        elif hint == 3:
            valid_version = "3.30"
        else:
            valid_version = "4.10"
        message = 'Unknown MDF version "{}". The available versions are {}; automatically using version "{}"'
        message = message.format(version, SUPPORTED_VERSIONS, valid_version)
        logger.warning(message)
    else:
        valid_version = typing.cast(v3c.Version2 | v3c.Version | v4c.Version, version)
    return valid_version


class ChannelsDB(dict[str, tuple[tuple[int, int], ...]]):
    def __init__(self) -> None:
        super().__init__()

    def add(self, channel_name: str, entry: tuple[int, int]) -> None:
        """Add name to channels database and check if it contains a source
        path.

        Parameters
        ----------
        channel_name : str
            Name that needs to be added to the database.
        entry : tuple
            (group index, channel index) pair.
        """
        if channel_name:
            if channel_name not in self:
                self[channel_name] = (entry,)
            else:
                self[channel_name] += (entry,)

            if "\\" in channel_name:
                channel_name, _ = channel_name.split("\\", 1)

                if channel_name not in self:
                    self[channel_name] = (entry,)
                elif entry not in self[channel_name]:
                    self[channel_name] += (entry,)


def randomized_string(size: int) -> bytes:
    """Get a null-terminated string of length `size`.

    Parameters
    ----------
    size : int
        Target string length.

    Returns
    -------
    string : bytes
        Randomized string.
    """
    return bytes(randint(65, 90) for _ in range(size - 1)) + b"\0"


def is_file_like(obj: object) -> TypeIs[FileLike]:
    """Check if the object is a file-like object.

    For objects to be considered file-like, they must
    be an iterator AND have a 'read' and 'seek' method
    as an attribute.

    Note: file-like objects must be iterable, but
    iterable objects need not be file-like.

    Parameters
    ----------
    obj : object
        The object to check.

    Returns
    -------
    is_file_like : bool
        Whether `obj` has file-like properties.

    Examples
    --------
    >>> buffer = BytesIO(b"data")
    >>> is_file_like(buffer)
    True
    >>> is_file_like([1, 2, 3])
    False
    """
    return isinstance(obj, FileLike) or isinstance(getattr(obj, "file", None), FileLike)


class UniqueDB:
    def __init__(self) -> None:
        self._db: dict[str, int] = {}

    def get_unique_name(self, name: str) -> str:
        """Return an available unique name.

        Parameters
        ----------
        name : str
            Name to be made unique.

        Returns
        -------
        unique_name : str
            New unique name.
        """

        if name not in self._db:
            self._db[name] = 0
            return name
        else:
            index = self._db[name]
            self._db[name] = index + 1
            return f"{name}_{index}"


def cut_video_stream(stream: bytes, start: float, end: float, fmt: str) -> bytes:
    """Cut video stream from `start` to `end` time.

    Parameters
    ----------
    stream : bytes
        Video file content.
    start : float
        Start time.
    end : float
        End time.

    Returns
    -------
    result : bytes
        Content of cut video.
    """
    with TemporaryDirectory() as tmp:
        in_file = Path(tmp) / f"in{fmt}"
        out_file = Path(tmp) / f"out{fmt}"

        in_file.write_bytes(stream)

        try:
            ret = subprocess.run(
                [
                    "ffmpeg",
                    "-ss",
                    f"{start}",
                    "-i",
                    f"{in_file}",
                    "-to",
                    f"{end}",
                    "-c",
                    "copy",
                    f"{out_file}",
                ],
                capture_output=True,
                check=False,
            )
        except FileNotFoundError:
            result = stream
        else:
            if ret.returncode:
                result = stream
            else:
                result = out_file.read_bytes()

    return result


def get_video_stream_duration(stream: bytes) -> float | None:
    with TemporaryDirectory() as tmp:
        in_file = Path(tmp) / "in"
        in_file.write_bytes(stream)

        try:
            process = subprocess.run(
                [
                    "ffprobe",
                    "-v",
                    "error",
                    "-show_entries",
                    "format=duration",
                    "-of",
                    "default=noprint_wrappers=1:nokey=1",
                    f"{in_file}",
                ],
                capture_output=True,
                check=False,
            )
            result = float(process.stdout)
        except FileNotFoundError:
            result = None
    return result


class VirtualChannelGroup:
    """Starting with MDF v4.20 it is possible to use remote masters and column
    oriented storage. This means we now have virtual channel groups that can
    span over multiple regular channel groups. This class facilitates the
    handling of these virtual groups.
    """

    __slots__ = (
        "cycles_nr",
        "groups",
        "record_size",
    )

    def __init__(self) -> None:
        self.groups: list[int] = []
        self.record_size = 0
        self.cycles_nr = 0

    def __repr__(self) -> str:
        return f"VirtualChannelGroup(groups={self.groups}, records_size={self.record_size}, cycles_nr={self.cycles_nr})"


def block_fields(obj: object) -> list[str]:
    fields = []
    for attr in dir(obj):
        if attr[:2] + attr[-2:] == "____":
            continue
        try:
            if callable(getattr(obj, attr)):
                continue
            fields.append(f"{attr}:{getattr(obj, attr)}")
        except AttributeError:
            continue

    return fields


@overload
def components(
    channel: NDArray[Any],
    channel_name: str,
    unique_names: UniqueDB,
    prefix: str = ...,
    master: Union[NDArray[Any], "pd.Index[Any]"] | None = ...,
    only_basenames: bool = ...,
    use_polars: Literal[False] = ...,
) -> Iterator[tuple[str, "pd.Series[Any]"]]: ...


@overload
def components(
    channel: NDArray[Any],
    channel_name: str,
    unique_names: UniqueDB,
    prefix: str = ...,
    master: Union[NDArray[Any], "pd.Index[Any]"] | None = ...,
    only_basenames: bool = ...,
    *,
    use_polars: Literal[True],
) -> Iterator[tuple[str, "list[object]"]]: ...


def components(
    channel: NDArray[Any],
    channel_name: str,
    unique_names: UniqueDB,
    prefix: str = "",
    master: Union[NDArray[Any], "pd.Index[Any]"] | None = None,
    only_basenames: bool = False,
    use_polars: bool = False,
) -> Iterator[tuple[str, Union["pd.Series[Any]", list[object]]]]:
    """Yield pandas Series and unique name based on the ndarray object.

    Parameters
    ----------
    channel : np.ndarray
        Channel to be used for Series.
    channel_name : str
        Channel name.
    unique_names : UniqueDB
        Unique names object.
    prefix : str, optional
        Prefix used in case of nested recarrays.
    master : np.ndarray | pd.Index, optional
        Optional index for the Series.
    only_basenames : bool, default False
        Use just the field names, without prefix, for structures and channel
        arrays.

        .. versionadded:: 5.13.0

    use_polars : bool, default False
        Use polars.

        .. versionadded:: 8.1.0

    Yields
    ------
    name, series : (str, values)
        Tuple of unique name and values.
    """
    names = channel.dtype.names

    # channel arrays
    if names and names[0] == channel_name:
        name = names[0]

        if not only_basenames:
            if prefix:
                name_ = unique_names.get_unique_name(f"{prefix}.{name}")
            else:
                name_ = unique_names.get_unique_name(name)
        else:
            name_ = unique_names.get_unique_name(name)

        samples = channel[name]
        if samples.dtype.byteorder not in target_byte_order:
            samples = samples.byteswap().view(samples.dtype.newbyteorder())

        if len(samples.shape) > 1:
            values = (
                list(samples)
                if use_polars
                else Series(
                    list(samples),
                    index=master,
                )
            )
        elif not use_polars:
            values = Series(
                samples,
                index=master,
            )

        yield name_, values

        for name in names[1:] if names else ():
            samples = channel[name]

            if samples.dtype.byteorder not in target_byte_order:
                samples = samples.byteswap().view(samples.dtype.newbyteorder())

            if not only_basenames:
                axis_name = unique_names.get_unique_name(f"{name_}.{name}")
            else:
                axis_name = unique_names.get_unique_name(name)
            if len(samples.shape) > 1:
                values = (
                    list(samples)
                    if use_polars
                    else Series(
                        list(samples),
                        index=master,
                    )
                )
            elif not use_polars:
                values = Series(
                    samples,
                    index=master,
                )

            yield axis_name, values

    # structure composition
    else:
        for name in channel.dtype.names or ():
            samples = channel[name]

            if samples.dtype.names:
                yield from components(
                    samples,
                    name,
                    unique_names,
                    prefix=f"{prefix}.{channel_name}" if prefix else f"{channel_name}",
                    master=master,
                    only_basenames=only_basenames,
                )

            else:
                if samples.dtype.byteorder not in target_byte_order:
                    samples = samples.byteswap().view(samples.dtype.newbyteorder())

                if not only_basenames:
                    name_ = unique_names.get_unique_name(
                        f"{prefix}.{channel_name}.{name}" if prefix else f"{channel_name}.{name}"
                    )
                else:
                    name_ = unique_names.get_unique_name(name)
                if len(samples.shape) > 1:
                    values = (
                        list(samples)
                        if use_polars
                        else Series(
                            list(samples),
                            index=master,
                        )
                    )
                elif not use_polars:
                    values = Series(
                        samples,
                        index=master,
                    )

                yield name_, values


class DataBlockInfo:
    __slots__ = (
        "address",
        "block_limit",
        "block_type",
        "compressed_size",
        "first_timestamp",
        "invalidation_block",
        "last_timestamp",
        "original_size",
        "param",
    )

    def __init__(
        self,
        address: int,
        block_type: int,
        original_size: int | None,
        compressed_size: int | None,
        param: int | None,
        invalidation_block: Optional["InvalidationBlockInfo"] = None,
        block_limit: int | None = None,
        first_timestamp: bytes | None = None,
        last_timestamp: bytes | None = None,
    ) -> None:
        self.address = address
        self.block_type = block_type
        self.original_size = original_size
        self.compressed_size = compressed_size
        self.param = param
        self.invalidation_block = invalidation_block
        self.block_limit = block_limit
        self.first_timestamp = first_timestamp
        self.last_timestamp = last_timestamp

    def __repr__(self) -> str:
        return (
            f"DataBlockInfo(address=0x{self.address:X}, "
            f"block_type={self.block_type}, "
            f"original_size={self.original_size}, "
            f"compressed_size={self.compressed_size}, "
            f"param={self.param}, "
            f"invalidation_block={self.invalidation_block}, "
            f"block_limit={self.block_limit}, "
            f"first_timestamp={self.first_timestamp!r}, "
            f"last_timestamp={self.last_timestamp!r})"
        )


class Fragment:
    def __init__(
        self,
        data: bytes | bytearray,
        record_offset: int = -1,
        record_count: int = -1,
        invalidation_data: bytes | bytearray | None = None,
        is_record: bool = True,
    ) -> None:
        self.data = data
        self.record_count = record_count
        self.record_offset = record_offset
        self.invalidation_data = invalidation_data
        self.is_record = is_record

    def __repr__(self) -> str:
        return (
            f"FragmentInfo({len(self.data)} bytes, "
            f"record_offset={self.record_offset}, "
            f"record_count={self.record_count}, "
            f"is_record={self.is_record})"
        )


class InvalidationBlockInfo(DataBlockInfo):
    __slots__ = ("all_valid",)

    def __init__(
        self,
        address: int,
        block_type: int,
        original_size: int | None,
        compressed_size: int | None,
        param: int | None,
        all_valid: bool = False,
        block_limit: int | None = None,
    ) -> None:
        super().__init__(address, block_type, original_size, compressed_size, param, block_limit=block_limit)
        self.all_valid = all_valid

    def __repr__(self) -> str:
        return (
            f"InvalidationBlockInfo(address=0x{self.address:X}, "
            f"block_type={self.block_type}, "
            f"original_size={self.original_size}, "
            f"compressed_size={self.compressed_size}, "
            f"param={self.param}, "
            f"all_valid={self.all_valid}, "
            f"block_limit={self.block_limit})"
        )


class SignalDataBlockInfo:
    __slots__ = (
        "address",
        "block_type",
        "compressed_size",
        "location",
        "original_size",
        "param",
    )

    def __init__(
        self,
        address: int,
        original_size: int,
        block_type: int = v4c.DT_BLOCK,
        param: int = 0,
        compressed_size: int | None = None,
        location: int = v4c.LOCATION_ORIGINAL_FILE,
    ) -> None:
        self.address = address
        self.compressed_size = compressed_size or original_size
        self.block_type = block_type
        self.original_size = original_size
        self.param = param
        self.location = location

    def __repr__(self) -> str:
        return (
            f"SignalDataBlockInfo(address=0x{self.address:X}, "
            f"original_size={self.original_size}, "
            f"compressed_size={self.compressed_size}, "
            f"block_type={self.block_type})"
        )


def get_fields(obj: object) -> list[str]:
    fields: list[str] = []
    for attr in dir(obj):
        if attr[:2] + attr[-2:] == "____":
            continue
        try:
            if callable(getattr(obj, attr)):
                continue
            fields.append(attr)
        except AttributeError:
            continue
    return fields


# code snippet taken from https://www.kaggle.com/arjanso/reducing-dataframe-memory-size-by-65
def downcast(array: NDArray[Any]) -> NDArray[Any]:
    kind = array.dtype.kind
    if kind == "f":
        array = array.astype(np.float32)
    elif kind in "ui":
        min_ = array.min()
        max_ = array.max()
        if min_ >= 0:
            if max_ < 255:
                array = array.astype(np.uint8)
            elif max_ < 65535:
                array = array.astype(np.uint16)
            elif max_ < 4294967295:
                array = array.astype(np.uint32)
            else:
                array = array.astype(np.uint64)
        else:
            if min_ > np.iinfo(np.int8).min and max_ < np.iinfo(np.int8).max:
                array = array.astype(np.int8)
            elif min_ > np.iinfo(np.int16).min and max_ < np.iinfo(np.int16).max:
                array = array.astype(np.int16)
            elif min_ > np.iinfo(np.int32).min and max_ < np.iinfo(np.int32).max:
                array = array.astype(np.int32)
            elif min_ > np.iinfo(np.int64).min and max_ < np.iinfo(np.int64).max:
                array = array.astype(np.int64)

    return array


def _csv_int2bin(val: int) -> str:
    """Format CAN id as bin.

    100 -> 1100100
    """

    return f"{val:b}"


csv_int2bin = np.vectorize(_csv_int2bin, otypes=[str])


def _csv_int2hex(val: "pd.Series[int]") -> str:
    """Format CAN id as hex.

    100 -> 64
    """

    return f"{val:X}"


csv_int2hex = np.vectorize(_csv_int2hex, otypes=[str])


def _csv_bytearray2hex(val: NDArray[Any], size: int | None = None) -> str:
    """Format CAN payload as hex strings.

    b'\xa2\xc3\x08' -> A2 C3 08
    """
    if size is not None:
        hex_val = typing.cast(bytes, val.tobytes())[:size].hex(" ", 1).upper()  # type: ignore[redundant-cast, unused-ignore]
    else:
        try:
            hex_val = typing.cast(bytes, val.tobytes()).hex(" ", 1).upper()  # type: ignore[redundant-cast, unused-ignore]
        except:
            hex_val = "●"

    return hex_val


csv_bytearray2hex = np.vectorize(_csv_bytearray2hex, otypes=[str])


def pandas_query_compatible(name: str) -> str:
    """Adjust column name for usage in DataFrame query string."""

    for c in ".$[]: ":
        name = name.replace(c, "_")

    if name.startswith(tuple(string.digits)):
        name = "file_" + name
    try:
        exec(f"from pandas import {name}")
    except ImportError:
        pass
    else:
        name = f"{name}__"
    return name


class _Kwargs(TypedDict, total=False):
    fd: bool
    load_flat: bool
    cluster_name: str


def load_can_database(
    path: StrPath, contents: bytes | str | None = None, **kwargs: Unpack[_Kwargs]
) -> CanMatrix | None:
    """

    Parameters
    ----------
    path : str | path-like
        Database path.
    contents : bytes | str, optional
        Optional database content.
    fd : bool, optional
        If supplied, only buses with the same FD kind will be loaded.
    load_flat : bool, optional
        If True, all the CAN messages found in multiple buses will be contained
        in the CAN database object. By default the first bus will be returned.
    cluster_name : str, optional
        If supplied, load just the clusters with this name.

    Returns
    -------
    db : canmatrix.CanMatrix | None
        CAN database object or None.
    """
    path = Path(path)
    import_type = path.suffix.lstrip(".").lower()

    try:
        if contents is None:
            dbs = canmatrix.formats.loadp(str(path), import_type=import_type, key="db", **kwargs)
        else:
            dbs = canmatrix.formats.loads(contents, import_type=import_type, key="db", **kwargs)
    except UnicodeDecodeError:
        if contents is None:
            contents = path.read_bytes()

        encoding = detect(contents)["encoding"]

        if encoding:
            try:
                dbs = canmatrix.formats.loads(
                    contents,
                    import_type=import_type,
                    key="db",
                    encoding=encoding,
                    **kwargs,
                )
            except:
                dbs = None

    if dbs:
        # filter only CAN clusters
        dbs = {name: db for name, db in dbs.items() if db.type == matrix_class.CAN}

    if dbs:
        cluster_name = kwargs.get("cluster_name", None)
        if cluster_name is not None:
            dbs = {name: db for name, db in dbs.items() if name == cluster_name}

        if "fd" in kwargs:
            fd = kwargs["fd"]
            dbs = {name: db for name, db in dbs.items() if db.contains_fd == fd}

        if kwargs.get("load_flat", False):
            can_matrix, *rest = list(dbs.values())
            can_matrix.merge(rest)

        else:
            first_bus = list(dbs)[0]
            can_matrix = dbs[first_bus]
    else:
        can_matrix = None

    return can_matrix


def all_blocks_addresses(obj: FileLike | mmap.mmap) -> tuple[dict[int, bytes], dict[bytes, list[int]], list[int]]:
    DG = "DG\x00\x00\x00\x00\x40\x00\x00\x00\x00\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00"
    others = "(D[VTZIL]|AT|C[AGHNC]|EV|FH|HL|LD|MD|R[DVI]|S[IRD]|TX|GD)\x00\x00\x00\x00"
    pattern = re.compile(
        f"(?P<block>##({DG}|{others}))".encode("ascii"),
        re.DOTALL | re.MULTILINE,
    )

    try:
        obj.seek(0)
    except:
        pass

    source: Buffer | bytes
    if isinstance(obj, Buffer):
        re.search(pattern, obj)
        source = obj
    else:
        source = obj.read()

    addresses: list[int] = []
    block_groups: dict[bytes, list[int]] = {}
    blocks: dict[int, bytes] = {}

    for match in re.finditer(pattern, source):
        btype: bytes = match.group("block")[:4]
        start = match.start()

        if start % 8:
            continue

        btype_addresses = block_groups.setdefault(btype, [])
        btype_addresses.append(start)
        addresses.append(start)
        blocks[start] = btype

    return blocks, block_groups, addresses


def plausible_timestamps(
    t: NDArray[Any],
    minimum: float,
    maximum: float,
    exp_min: int = -15,
    exp_max: int = 15,
) -> tuple[bool, NDArray[bool_]]:
    """Check if the timestamps are plausible.

    Parameters
    ----------
    t : np.ndarray
        Timestamps array.
    minimum : float
        Minimum plausible timestamp.
    maximum : float
        Maximum plausible timestamp.
    exp_min : int, default -15
        Minimum plausible exponent used for the timestamps float values.
    exp_max : int, default 15
        Maximum plausible exponent used for the timestamps float values.

    Returns
    -------
    all_ok, idx : (bool, np.ndarray)
        The `all_ok` flag to indicate if all the timestamps are ok; this can
        be checked before applying the indexing array.
    """

    exps = np.log10(t)
    idx = (~np.isnan(t)) & (~np.isinf(t)) & (t >= minimum) & (t <= maximum) & (t == 0) | (
        (exps >= exp_min) & (exps <= exp_max)
    )
    if not np.all(idx):
        all_ok = False
        return all_ok, idx
    else:
        all_ok = True
        return all_ok, idx


table = str.maketrans(
    {
        "<": "&lt;",
        ">": "&gt;",
        "&": "&amp;",
        "'": "&apos;",
        '"': "&quot;",
    }
)


def escape_xml_string(string: str) -> str:
    return string.translate(table)


class SignalFlags:
    no_flags = 0x0
    user_defined_comment = 0x1
    user_defined_conversion = 0x2
    user_defined_unit = 0x4
    user_defined_name = 0x8
    stream_sync = 0x10
    computed = 0x20
    virtual = 0x40
    virtual_master = 0x80


_Params = ParamSpec("_Params")
_Ret = TypeVar("_Ret")


def timeit(func: Callable[_Params, _Ret]) -> Callable[_Params, _Ret]:
    def timed(*args: _Params.args, **kwargs: _Params.kwargs) -> _Ret:
        t1 = perf_counter()
        ret = func(*args, **kwargs)
        t2 = perf_counter()
        delta = t2 - t1
        if delta >= 1e-3:
            print(f"CALL {func.__qualname__}: {delta * 1e3:.3f} ms")
        else:
            print(f"CALL {func.__qualname__}: {delta * 1e6:.3f} us")
        return ret

    return timed


class Timer:
    def __init__(self, name: str = "") -> None:
        self.name = name or str(id(self))
        self.count = 0
        self.total_time = 0.0

    def __enter__(self) -> "Timer":
        now = perf_counter()
        self.start = now
        return self

    def __exit__(
        self, type: type[BaseException] | None, value: BaseException | None, traceback: TracebackType | None
    ) -> None:
        now = perf_counter()
        self.total_time += now - self.start
        self.count += 1

    def display(self) -> None:
        if self.count:
            for factor, r, unit in ((1e3, 3, "ms"), (1e6, 6, "us"), (1e9, 9, "ns")):
                tpi = round(self.total_time / self.count, r)
                if tpi:
                    break
            print(
                f"""TIMER {self.name}:
\t* {self.count} iterations in {self.total_time * 1000:.3f}ms
\t* {self.count / self.total_time:.3f} iter/s
\t* {self.total_time / self.count * factor:.3f} {unit}/iter"""
            )
        else:
            print(f"TIMER {self.name}:\n\t* inactive")


def handle_incomplete_block(address: int, file: StrPath | None = None) -> None:
    if file is None:
        file = "The file"
    else:
        file = Path(file).name

    msg = f"Incomplete block at 0x{address:x} exceeds the file size. {file} might be corrupted or partially written."
    logger.warning(msg)

    if GLOBAL_OPTIONS["raise_on_incomplete_blocks"]:
        raise MdfException(msg)



#!/usr/bin/env python
import json
from json import JSONDecodeError
import pathlib
import re
from unittest import mock
from unittest.mock import ANY

from PySide6 import QtCore, QtGui, QtTest, QtWidgets

from test.asammdf.gui.test_base import safe_setup
from test.asammdf.gui.widgets.test_BasePlotWidget import TestPlotWidget


class TestContextMenu(TestPlotWidget):
    # Note: Test Plot Widget through FileWidget.

    @safe_setup
    def setUp(self):
        super().setUp()

        self.setUpFileWidget(measurement_file=self.measurement_file, default=True)
        # Switch ComboBox to "Natural sort"
        self.widget.channel_view.setCurrentText("Natural sort")

        self.create_window(window_type="Plot", channels_indexes=[10, 11, 12, 13, 15])
        self.assertEqual(len(self.widget.mdi_area.subWindowList()), 1)

        # Drag and Drop channel from FileWidget.channel_tree to Plot
        self.plot = self.widget.mdi_area.subWindowList()[0].widget()
        self.plot_channel_a = self.plot.channel_selection.topLevelItem(0)
        self.plot_channel_a.setCheckState(self.Column.NAME, QtCore.Qt.CheckState.Checked)
        self.plot_channel_b = self.plot.channel_selection.topLevelItem(1)
        self.plot_channel_b.setCheckState(self.Column.NAME, QtCore.Qt.CheckState.Checked)
        self.plot_channel_c = self.plot.channel_selection.topLevelItem(2)
        self.plot_channel_c.setCheckState(self.Column.NAME, QtCore.Qt.CheckState.Checked)
        self.plot_channel_d = self.plot.channel_selection.topLevelItem(3)
        self.plot_channel_d.setCheckState(self.Column.NAME, QtCore.Qt.CheckState.Checked)
        self.plot_channel_e = self.plot.channel_selection.topLevelItem(4)
        self.plot_channel_e.setCheckState(self.Column.NAME, QtCore.Qt.CheckState.Checked)

        self.processEvents()

    def test_Action_SearchItem_Cancel(self):
        """
        Test Scope:
            Ensure that action 'Search Item' call 'QtWidgets.QInputDialog.getText' in order to get a channel name.
            The Dialog is canceled and nothing happens.
        Events:
            - Open Context Menu
            - Trigger 'Search Item' action
                - Simulate that the dialog is canceled.
        Evaluate:
            - Ensure that there is no selected channel because
            performing Right Click for ContextMenu will clear selection of the tree.
        """
        QtTest.QTest.keyClick(self.plot.channel_selection.viewport(), QtCore.Qt.Key.Key_Down)
        self.assertEqual(1, len(self.plot.channel_selection.selectedItems()))

        with mock.patch("asammdf.gui.widgets.tree.QtWidgets.QInputDialog.getText") as mo_getText:
            mo_getText.return_value = None, None
            self.context_menu(action_text="Search item")

        self.assertEqual(0, len(self.plot.channel_selection.selectedItems()))

    def test_Action_SearchItem_NonexistentChannel(self):
        """
        Test Scope:
            Ensure that action 'Search Item' call 'QtWidgets.QInputDialog.getText' in order to get a channel name.
            The Dialog is filled with non-existing channel name.
        Events:
            - Open Context Menu
            - Trigger 'Search Item' action
                - Simulate that the dialog is filled with non-existing channel name.
        Evaluate:
            - Ensure that selection is cleared.
            - Ensure that MessageBox informs user.
        """
        QtTest.QTest.keyClick(self.plot.channel_selection.viewport(), QtCore.Qt.Key.Key_Down)
        self.assertEqual(1, len(self.plot.channel_selection.selectedItems()))

        with (
            mock.patch("asammdf.gui.widgets.tree.QtWidgets.QInputDialog.getText") as mo_getText,
            mock.patch("asammdf.gui.widgets.tree.MessageBox.warning") as mo_warning,
        ):
            mo_getText.return_value = self.id(), True
            self.context_menu(action_text="Search item")

        self.assertEqual(0, len(self.plot.channel_selection.selectedItems()))
        mo_warning.assert_called_with(self.plot.channel_selection, "No matches found", ANY)

    def test_Action_SearchItem_ExistentChannel(self):
        """
        Test Scope:
            Ensure that action 'Search Item' call 'QtWidgets.QInputDialog.getText' in order to get a channel name.
            The Dialog is filled with existing channel name.
            Found channel is selected on tree
        Events:
            - Open Context Menu
            - Trigger 'Search Item' action
                - Simulate that the dialog is filled with existing channel name.
        Evaluate:
            - Ensure that selection is set on the found channel.
        """
        QtTest.QTest.keyClick(self.plot.channel_selection.viewport(), QtCore.Qt.Key.Key_Down)
        self.assertEqual(1, len(self.plot.channel_selection.selectedItems()))

        with (
            mock.patch("asammdf.gui.widgets.tree.QtWidgets.QInputDialog.getText") as mo_getText,
            mock.patch("asammdf.gui.widgets.tree.MessageBox.warning") as mo_warning,
        ):
            mo_getText.return_value = self.plot_channel_b.text(self.Column.NAME), True
            self.context_menu(action_text="Search item")

        self.assertEqual(1, len(self.plot.channel_selection.selectedItems()))
        mo_warning.assert_not_called()

    def test_Action_AddChannelGroup_Cancel(self):
        """
        Test Scope:
            - Ensure that action 'Add channel group' will add new item (type Group) on tree.
        Events:
            - Open Context Menu
            - Trigger 'Add channel group' action
                - Simulate that the dialog is canceled.
        """
        channels_nr = self.plot.channel_selection.topLevelItemCount()

        with mock.patch("asammdf.gui.widgets.tree.QtWidgets.QInputDialog.getText") as mo_getText:
            mo_getText.return_value = "FirstGroup", None
            self.context_menu(action_text="Add channel group")

        self.assertEqual(channels_nr, self.plot.channel_selection.topLevelItemCount())

    def test_Action_AddChannelGroup(self):
        """
        Test Scope:
            - Ensure that action 'Add channel group' will add new item (type Group) on tree.
        Events:
            - Open Context Menu
            - Trigger 'Add channel group' action
                - Simulate that the dialog is confirmed.
        Evaluate:
            - Check if there is one extra element in tree.
        """
        channels_nr = self.plot.channel_selection.topLevelItemCount()

        with mock.patch("asammdf.gui.widgets.tree.QtWidgets.QInputDialog.getText") as mo_getText:
            mo_getText.return_value = "FirstGroup", True
            self.context_menu(action_text="Add channel group")

        self.assertEqual(channels_nr + 1, self.plot.channel_selection.topLevelItemCount())

    def test_Action_CopyNames(self):
        """
        Test Scope:
            - Ensure that action 'Copy Names' will place channel names in clipboard.
        Events:
            - Select one channel
            - Open Context Menu
            - Trigger 'Copy Names' action
            - Select two channels
            - Open Context Menu
            - Trigger 'Copy Names' action
        Evaluate:
            - Evaluate that channels names are placed in clipboard.
        """

        with self.subTest("1Channel"):
            position = self.plot.channel_selection.visualItemRect(self.plot_channel_a).center()
            self.context_menu(action_text="Copy names", position=position)

            clipboard = QtWidgets.QApplication.instance().clipboard().text()
            self.assertEqual(self.plot_channel_a.text(self.Column.NAME), clipboard)

        with self.subTest("2Channels"):
            self.plot_channel_a.setSelected(True)
            self.plot_channel_b.setSelected(True)
            position_1 = self.plot.channel_selection.visualItemRect(self.plot_channel_b).center()
            self.context_menu(action_text="Copy names", position=position_1)

            clipboard = QtWidgets.QApplication.instance().clipboard().text()
            channels = (self.plot_channel_a.text(self.Column.NAME), self.plot_channel_b.text(self.Column.NAME))
            self.assertEqual("\n".join(channels), clipboard)

    def test_Action_CopyNamesAndValues(self):
        """
        Test Scope:
            - Ensure that action 'Copy names and value' will place channel names in clipboard.
        Events:
            - Select one channel
            - Open Context Menu
            - Trigger 'Copy names and value' action
            - Select two channels
            - Open Context Menu
            - Trigger 'Copy names and value' action
        Evaluate:
            - Evaluate that channels names are placed in clipboard.
        """

        with self.subTest("1Channel"):
            position = self.plot.channel_selection.visualItemRect(self.plot_channel_a).center()
            self.context_menu(action_text="Copy names and values", position=position)

            clipboard = QtWidgets.QApplication.instance().clipboard().text()
            pattern_name = re.escape(self.plot_channel_a.text(self.Column.NAME))
            pattern_unit = re.escape(self.plot_channel_a.text(self.Column.UNIT))
            self.assertRegex(clipboard, expected_regex=f"{pattern_name}, t = \\d+[.]?\\d+s, \\d+[.]?\\d+{pattern_unit}")

        with self.subTest("2Channels"):
            self.plot_channel_a.setSelected(True)
            self.plot_channel_b.setSelected(True)
            position_1 = self.plot.channel_selection.visualItemRect(self.plot_channel_b).center()
            self.context_menu(action_text="Copy names and values", position=position_1)

            clipboard = QtWidgets.QApplication.instance().clipboard().text()
            expected_regex = []
            for channel in (self.plot_channel_a, self.plot_channel_b):
                pattern_name = re.escape(channel.text(self.Column.NAME))
                pattern_unit = re.escape(channel.text(self.Column.UNIT))
                expected_regex.append(f"{pattern_name}, t = \\d+[.]?\\d+s, \\d+[.]?\\d+{pattern_unit}")

            self.assertRegex(clipboard, "\\n".join(expected_regex))

    def test_Action_CopyDisplayProperties_Channel(self):
        """
        Test Scope:
            - Ensure that channel display properties are copied in clipboard.
        Events:
            - Select one channel
            - Open Context Menu
            - Trigger 'Copy display properties' action
        Evaluate:
            - Evaluate that channel display properties are stored in clipboard in json format.
        """
        position = self.plot.channel_selection.visualItemRect(self.plot_channel_a).center()
        self.context_menu(action_text="Copy display properties", position=position)

        clipboard = QtWidgets.QApplication.instance().clipboard().text()
        try:
            content = json.loads(clipboard)
        except JSONDecodeError:
            self.fail("Clipboard Content cannot be decoded as JSON content.")
        else:
            self.assertIsInstance(content, dict)

    def test_Action_PasteDisplayProperties_Channel(self):
        """
        Test Scope:
            - Ensure that channel display properties can be pasted over different channel.
        Events:
            - Select one channel
            - Open Context Menu
            - Trigger 'Copy display properties' action
            - Select a different channel
            - Open Context Menu
            - Trigger 'Paste display properties' action
            - Select dst channel
            - Open Context Menu
            - Trigger 'Copy display properties' action
        Evaluate:
            - Evaluate that display properties are transferred from one channel to another
        """
        action_copy = "Copy display properties"
        action_paste = "Paste display properties"

        position_src = self.plot.channel_selection.visualItemRect(self.plot_channel_a).center()
        self.context_menu(action_text=action_copy, position=position_src)

        channel_a_properties = QtWidgets.QApplication.instance().clipboard().text()

        # Paste
        position_dst = self.plot.channel_selection.visualItemRect(self.plot_channel_b).center()
        self.context_menu(action_text=action_paste, position=position_dst)

        # Copy
        position_src = self.plot.channel_selection.visualItemRect(self.plot_channel_b).center()
        self.context_menu(action_text=action_copy, position=position_src)

        channel_b_properties = QtWidgets.QApplication.instance().clipboard().text()

        self.assertEqual(channel_a_properties, channel_b_properties)

    def test_Action_CopyChannelStructure_Channel(self):
        """
        Test Scope:
            - Ensure that channel structure is copied in clipboard.
        Events:
            - Select one channel
            - Open Context Menu
            - Trigger 'Copy channel structure' action
        Evaluate:
            - Evaluate that channel structure is stored in clipboard in json format.
        """
        position = self.plot.channel_selection.visualItemRect(self.plot_channel_a).center()
        self.context_menu(action_text="Copy channel structure", position=position)

        clipboard = QtWidgets.QApplication.instance().clipboard().text()
        try:
            content = json.loads(clipboard)
        except JSONDecodeError:
            self.fail("Clipboard Content cannot be decoded as JSON content.")
        else:
            self.assertIsInstance(content, list)
            for channel_properties in content:
                self.assertIsInstance(channel_properties, dict)

    def test_Action_PasteChannelStructure_Channel(self):
        """
        Test Scope:
            - Ensure that channel and structure is duplicated and structure is kept.
        Events:
            - Select one channel
            - Open Context Menu
            - Trigger 'Copy display properties' action
            - Trigger 'Copy channel structure' action
            - Open Context Menu
            - Trigger 'Paste channel structure' action
            - Select dst channel
            - Open Context Menu
            - Trigger 'Copy display properties' action
        Evaluate:
            - Evaluate that channel is duplicated and structure is kept.
        """
        action_copy_dsp_properties = "Copy display properties"
        action_copy = "Copy channel structure"
        action_paste = "Paste channel structure"

        position_src = self.plot.channel_selection.visualItemRect(self.plot_channel_a).center()
        # Copy Channel Structure
        self.context_menu(action_text=action_copy, position=position_src)

        channels_count = self.plot.channel_selection.topLevelItemCount()

        # Paste Channel Structure
        self.context_menu(action_text=action_paste)

        self.assertEqual(channels_count + 1, self.plot.channel_selection.topLevelItemCount())

        channels = self.plot.channel_selection.findItems(
            self.plot_channel_a.text(self.Column.NAME), QtCore.Qt.MatchFlags()
        )
        self.assertEqual(2, len(channels))

        # Copy DSP Properties
        channel_properties = []
        for channel in channels:
            position_src = self.plot.channel_selection.visualItemRect(channel).center()
            self.context_menu(action_text=action_copy_dsp_properties, position=position_src)
            channel_properties.append(QtWidgets.QApplication.instance().clipboard().text())

        self.assertEqual(channel_properties[0], channel_properties[1])

        self.processEvents(0.2)

    def test_Menu_EnableDisable_Action_DeactivateGroups_Channel(self):
        """
        Test Scope:
            - Ensure that group items can be disabled from Context Menu.
        Events:
            - Select 1 channel
            - Open Context Menu
            - Trigger "Enable/Disable -> Deactivate Groups" action
        Expected:
            - Individual channels that are not part of group should not be affected
        """
        # Events
        position = self.plot.channel_selection.visualItemRect(self.plot_channel_a).center()
        self.context_menu(action_text="Deactivate groups", position=position)

        # Evaluate
        self.assertEqual(False, self.plot_channel_a.isDisabled())

    def test_Menu_EnableDisable_Action_DisableAllButThis(self):
        """
        Test Scope:
            - Ensure that all channels are disabled except the item selected.
        Events:
            - Open Context Menu
            - Select action "Disable all but this" from sub-menu "Enable/disable"
        Expected:
            - Evaluate that all channels are disabled except selected item.
        """
        with mock.patch("asammdf.gui.widgets.tree.QtWidgets.QInputDialog.getText") as mo_getText:
            mo_getText.return_value = "A", True
            self.context_menu(action_text="Add channel group")

        positions_src = self.plot.channel_selection.visualItemRect(self.plot_channel_b).center()
        self.context_menu(action_text="Disable all but this", position=positions_src)

        # Evaluate
        self.assertEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_b.checkState(self.Column.NAME))
        count = self.plot.channel_selection.topLevelItemCount()
        for i in range(count):
            item = self.plot.channel_selection.topLevelItem(i)
            if item.type() != item.Info and item != self.plot_channel_b:
                self.assertEqual(QtCore.Qt.CheckState.Unchecked, item.checkState(self.Column.NAME))

    def test_Menu_ShowHide_Action_HideDisabledItems(self):
        """
        Test Scope:
            - Ensure that item is hidden from channel selection when is disabled.
        Events:
            - Disable 1 channel by key Space
            - Disable 1 channel by mouseClick on item CheckBox
        Evaluate:
            - Evaluate that items that are unchecked are not present anymore on channel selection
        """
        self.context_menu(action_text="Hide disabled items")

        with self.subTest("DisableBySpace"):
            # Select one channel
            self.mouseClick_WidgetItem(self.plot_channel_a)
            # Event
            QtTest.QTest.keyClick(self.plot.channel_selection, QtCore.Qt.Key.Key_Space)
            # Evaluate
            self.assertTrue(self.plot_channel_a.isHidden())

        with self.subTest("DisableByClick"):
            pos = self.plot.channel_selection.visualItemRect(self.plot_channel_b).center()
            # Magic Number to detect center of checkbox
            pos = QtCore.QPoint(28, pos.y())
            # Event
            QtTest.QTest.mouseClick(
                self.plot.channel_selection.viewport(),
                QtCore.Qt.MouseButton.LeftButton,
                QtCore.Qt.KeyboardModifier.NoModifier,
                pos,
            )
            # Evaluate
            self.assertTrue(self.plot_channel_b.isHidden())

    def test_Menu_ShowHide_Action_ShowDisabledItems(self):
        """
        Test Scope:
            - Ensure that item is showed on channel selection when is disabled.
        Events:
            - Disable 1 channel by key Space
            - Disable 1 channel by mouseClick on item CheckBox
        Evaluate:
            - Evaluate that items that are unchecked are not present anymore on channel selection
        """

        with self.subTest("DisableBySpace"):
            self.context_menu(action_text="Hide disabled items")
            # Select one channel
            self.mouseClick_WidgetItem(self.plot_channel_a)
            # Event
            QtTest.QTest.keyClick(self.plot.channel_selection, QtCore.Qt.Key.Key_Space)
            # Evaluate
            self.assertTrue(self.plot_channel_a.isHidden())
            self.context_menu(action_text="Hide disabled items")
            self.assertFalse(self.plot_channel_a.isHidden())

        with self.subTest("DisableByClick"):
            self.context_menu(action_text="Hide disabled items")
            pos = self.plot.channel_selection.visualItemRect(self.plot_channel_b).center()
            # Magic Number to detect center of checkbox
            pos = QtCore.QPoint(28, pos.y())
            # Event
            QtTest.QTest.mouseClick(
                self.plot.channel_selection.viewport(),
                QtCore.Qt.MouseButton.LeftButton,
                QtCore.Qt.KeyboardModifier.NoModifier,
                pos,
            )
            # Evaluate
            self.assertTrue(self.plot_channel_b.isHidden())
            self.context_menu(action_text="Hide disabled items")
            self.assertFalse(self.plot_channel_b.isHidden())

    def test_Menu_ShowHide_Action_HideMissingItems(self):
        """
        Test Scope:
            - Ensure that missing item is hidden from channel selection.
        Events:
            - Disable 1 channel by key Space
            - Disable 1 channel by mouseClick on item CheckBox
        Evaluate:
            - Evaluate that missing items are not present anymore on channel selection
        """
        dspf_filepath = pathlib.Path(self.resource, "missingItems.dspf")
        self.load_display_file(display_file=dspf_filepath)
        self.plot = self.widget.mdi_area.subWindowList()[0].widget()

        plot_channel = self.find_channel(channel_tree=self.plot.channel_selection, channel_name="1stMissingItem")
        self.processEvents()

        self.assertFalse(plot_channel.isHidden())

        # Events
        self.context_menu(action_text="Hide missing items")

        # Evaluate
        self.assertTrue(plot_channel.isHidden())
        self.processEvents(timeout=0.01)

    def test_Menu_ShowHide_Action_ShowMissingItems(self):
        """
        Test Scope:
            - Ensure that missing item is visible from channel selection.
        Events:
            - Open Context Menu
            - Select action: Hide missing items
            - Disable 1 channel by key Space
            - Disable 1 channel by mouseClick on item CheckBox
        Evaluate:
            - Evaluate that missing items are present anymore on channel selection
        """
        dspf_filepath = pathlib.Path(self.resource, "missingItems.dspf")
        self.load_display_file(display_file=dspf_filepath)
        self.plot = self.widget.mdi_area.subWindowList()[0].widget()

        self.context_menu(action_text="Hide missing items")

        plot_channel = self.find_channel(channel_tree=self.plot.channel_selection, channel_name="1stMissingItem")
        self.processEvents()

        self.assertTrue(plot_channel.isHidden())

        # Events
        self.context_menu(action_text="Hide missing items")

        # Evaluate
        self.assertFalse(plot_channel.isHidden())
        self.processEvents(timeout=0.01)

    def test_Menu_ShowHide_Action_FilterOnlyComputedChannels(self):
        """
        Test Scope:
            - Ensure that all channels are hidden from channel selection except VirtualChannels.
        Events:
            - Open Context Menu
            - Select Filter Only Computed Channels
        Evaluate:
            - Evaluate that channels items are not present anymore on channel selection except VirtualChannels.
        """
        dspf_filepath = pathlib.Path(self.resource, "missingItems.dspf")
        self.load_display_file(display_file=dspf_filepath)
        self.plot = self.widget.mdi_area.subWindowList()[0].widget()

        iterator = QtWidgets.QTreeWidgetItemIterator(self.plot.channel_selection)
        while iterator.value():
            item = iterator.value()
            if item:
                self.assertFalse(item.isHidden())
            iterator += 1

        # Events
        self.context_menu(action_text="Filter only computed channels")

        # Evaluate
        iterator = QtWidgets.QTreeWidgetItemIterator(self.plot.channel_selection)
        while iterator.value():
            item = iterator.value()
            if item and item.text(0) == "FirstVirtualChannel":
                self.assertFalse(item.isHidden())
            else:
                self.assertTrue(item.isHidden())
            iterator += 1

    def test_Menu_ShowHide_Action_UnfilterComputedChannels(self):
        """
        Test Scope:
            - Ensure that all channels are hidden from channel selection except VirtualChannels.
        Events:
            - Open Context Menu
            - Select Filter Only Computed Channels
        Evaluate:
            - Evaluate that channels items are not present anymore on channel selection except VirtualChannels.
        """
        dspf_filepath = pathlib.Path(self.resource, "missingItems.dspf")
        self.load_display_file(display_file=dspf_filepath)
        self.plot = self.widget.mdi_area.subWindowList()[0].widget()

        # Events
        self.context_menu(action_text="Filter only computed channels")
        # Evaluate
        iterator = QtWidgets.QTreeWidgetItemIterator(self.plot.channel_selection)
        while iterator.value():
            item = iterator.value()
            if item and item.text(0) == "FirstVirtualChannel":
                self.assertFalse(item.isHidden())
            else:
                self.assertTrue(item.isHidden())
            iterator += 1

        # Events
        self.context_menu(action_text="Filter only computed channels")

        # Evaluate
        iterator = QtWidgets.QTreeWidgetItemIterator(self.plot.channel_selection)
        while iterator.value():
            item = iterator.value()
            if item:
                self.assertFalse(item.isHidden())
            iterator += 1

    def test_Action_EditYAxisScaling(self):
        """
        Test Scope:
            - Ensure that action forwards the request to plot.
        Event:
            - Select channel
            - Open Context Menu
            - Select 'Edit Y axis scaling'
        Evaluate:
            - Evaluate that event is forwarded to plot
        """
        # Setup
        position = self.plot.channel_selection.visualItemRect(self.plot_channel_a).center()
        with mock.patch.object(self.plot, "keyPressEvent") as mo_keyPressEvent:
            self.context_menu(action_text="Edit Y axis scaling", position=position)
            mo_keyPressEvent.assert_called()
            event = mo_keyPressEvent.call_args.args[0]
            self.assertEqual(QtCore.QEvent.Type.KeyPress, event.type())
            self.assertEqual(QtCore.Qt.Key.Key_G, event.key())
            self.assertEqual(QtCore.Qt.KeyboardModifier.ControlModifier, event.modifiers())

    def test_Action_AddToCommonYAxis(self):
        """
        Test Scope:
            - Ensure that action will mark as checked checkbox for Y Axis.
        Event:
            - Select one channel
            - Open Context Menu
            - Select 'Add to common Y axis'
            - Select two channels
            - Open Context Menu
            - Select 'Add to common Y axis'
        Evaluate:
            - Evaluate that checkbox on column COMMON_AXIS is checked.
        """
        with self.subTest("1Channel"):
            self.assertNotEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_a.checkState(self.Column.COMMON_AXIS))

            position = self.plot.channel_selection.visualItemRect(self.plot_channel_a).center()
            self.context_menu(action_text="Add to common Y axis", position=position)

            self.assertEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_a.checkState(self.Column.COMMON_AXIS))

            self.context_menu(action_text="Add to common Y axis", position=position)

            self.assertEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_a.checkState(self.Column.COMMON_AXIS))

        with self.subTest("2Channels"):
            self.plot_channel_b.setSelected(True)
            self.plot_channel_c.setSelected(True)
            self.assertNotEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_b.checkState(self.Column.COMMON_AXIS))
            self.assertNotEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_c.checkState(self.Column.COMMON_AXIS))

            position = self.plot.channel_selection.visualItemRect(self.plot_channel_c).center()
            self.context_menu(action_text="Add to common Y axis", position=position)

            self.assertEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_b.checkState(self.Column.COMMON_AXIS))
            self.assertEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_c.checkState(self.Column.COMMON_AXIS))

            self.context_menu(action_text="Add to common Y axis", position=position)

            self.assertEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_b.checkState(self.Column.COMMON_AXIS))
            self.assertEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_c.checkState(self.Column.COMMON_AXIS))

    def test_Action_RemoveFromCommonYAxis(self):
        """
        Test Scope:
            - Ensure that action will mark as unchecked checkbox for Y Axis.
        Event:
            - Select one channel
            - Open Context Menu
            - Select 'Remove from common Y axis'
            - Select two channels
            - Open Context Menu
            - Select 'Remove from common Y axis'
        Evaluate:
            - Evaluate that checkbox on column COMMON_AXIS is unchecked.
        """
        with self.subTest("1Channel"):
            # Setup
            position = self.plot.channel_selection.visualItemRect(self.plot_channel_a).center()

            # Pre-evaluation
            self.assertNotEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_a.checkState(self.Column.COMMON_AXIS))

            # Event
            self.context_menu(action_text="Remove from common Y axis", position=position)
            # Evaluate
            self.assertNotEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_a.checkState(self.Column.COMMON_AXIS))

            # Event
            self.context_menu(action_text="Add to common Y axis", position=position)
            self.assertEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_a.checkState(self.Column.COMMON_AXIS))
            self.context_menu(action_text="Remove from common Y axis", position=position)
            # Evaluate
            self.assertNotEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_a.checkState(self.Column.COMMON_AXIS))

            # Event
            self.context_menu(action_text="Remove from common Y axis", position=position)
            # Evaluate
            self.assertNotEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_a.checkState(self.Column.COMMON_AXIS))

        with self.subTest("2Channels"):
            # Setup
            self.plot_channel_b.setSelected(True)
            self.plot_channel_c.setSelected(True)
            position_c = self.plot.channel_selection.visualItemRect(self.plot_channel_c).center()

            # Pre-evaluation
            self.assertNotEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_b.checkState(self.Column.COMMON_AXIS))
            self.assertNotEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_c.checkState(self.Column.COMMON_AXIS))

            # Event

            self.context_menu(action_text="Remove from common Y axis", position=position_c)
            # Evaluate
            self.assertNotEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_b.checkState(self.Column.COMMON_AXIS))
            self.assertNotEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_c.checkState(self.Column.COMMON_AXIS))

            # Event
            self.context_menu(action_text="Add to common Y axis", position=position_c)
            self.assertEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_b.checkState(self.Column.COMMON_AXIS))
            self.assertEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_c.checkState(self.Column.COMMON_AXIS))
            self.context_menu(action_text="Remove from common Y axis", position=position)
            # Evaluate
            self.assertNotEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_b.checkState(self.Column.COMMON_AXIS))
            self.assertNotEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_c.checkState(self.Column.COMMON_AXIS))

            # Event
            self.context_menu(action_text="Remove from common Y axis", position=position)
            # Evaluate
            self.assertNotEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_b.checkState(self.Column.COMMON_AXIS))
            self.assertNotEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_c.checkState(self.Column.COMMON_AXIS))

    def test_Action_SetColor(self):
        """
        Test Scope:
            - Ensure that channel color is changed.
        Events:
            - Open Context Menu
            - Select 'Set color [C]'
            - Select 1 Channel
            - Open Context Menu
            - Select 'Set color [C]'
            - Select 2 Channels
            - Open Context Menu
            - Select 'Set color'
        Evaluate:
            - Evaluate that color dialog is not open if channel is not selected.
            - Evaluate that channel color is changed.
        """
        action_text = "Set color"

        # Event
        with self.subTest("NoChannelSelected"):
            with mock.patch("asammdf.gui.widgets.tree.QtWidgets.QColorDialog.getColor") as mo_getColor:
                self.context_menu(action_text=action_text)
                mo_getColor.assert_not_called()

        with self.subTest("1ChannelSelected"):
            position = self.plot.channel_selection.visualItemRect(self.plot_channel_a).center()
            with mock.patch("asammdf.gui.widgets.tree.QtWidgets.QColorDialog.getColor") as mo_getColor:
                # Setup
                previous_color = self.plot_channel_a.color.name()
                color = QtGui.QColor("red")
                color_name = color.name()
                mo_getColor.return_value = color
                # Event
                self.context_menu(action_text=action_text, position=position)
                # Evaluate
                current_color = self.plot_channel_a.color.name()
                mo_getColor.assert_called()
                self.assertNotEqual(previous_color, current_color)
                self.assertEqual(color_name, current_color)

        with self.subTest("2ChannelsSelected"):
            self.mouseClick_WidgetItem(self.plot_channel_b)
            self.plot_channel_b.setSelected(True)
            self.plot_channel_c.setSelected(True)
            position = self.plot.channel_selection.visualItemRect(self.plot_channel_c).center()
            with mock.patch("asammdf.gui.widgets.tree.QtWidgets.QColorDialog.getColor") as mo_getColor:
                # Setup
                previous_b_color = self.plot_channel_b.color.name()
                previous_c_color = self.plot_channel_c.color.name()
                color = QtGui.QColor("blue")
                color_name = color.name()
                mo_getColor.return_value = color
                # Event
                self.context_menu(action_text=action_text, position=position)
                # Evaluate
                current_b_color = self.plot_channel_b.color.name()
                current_c_color = self.plot_channel_c.color.name()
                mo_getColor.assert_called()
                self.assertNotEqual(previous_b_color, current_b_color)
                self.assertNotEqual(previous_c_color, current_c_color)
                self.assertEqual(color_name, current_b_color)
                self.assertEqual(color_name, current_c_color)

    def test_Action_SetRandomColor(self):
        """
        Test Scope:
            - Ensure that channel color is changed.
        Events:
            - Open Context Menu
            - Select 'Set random color'
            - Select 1 Channel
            - Open Context Menu
            - Select 'Set random color'
            - Select 2 Channels
            - Open Context Menu
            - Select 'Set random color'
        Evaluate:
            - Evaluate that color dialog is not open if channel is not selected.
            - Evaluate that channel color is changed.
        """
        action_text = "Set random color"

        # Event
        with self.subTest("NoChannelSelected"):
            with mock.patch("asammdf.gui.widgets.tree.QtWidgets.QColorDialog.getColor") as mo_getColor:
                self.context_menu(action_text=action_text)
                mo_getColor.assert_not_called()

        with self.subTest("1ChannelSelected"):
            position = self.plot.channel_selection.visualItemRect(self.plot_channel_a).center()
            # Setup
            previous_color = self.plot_channel_a.color.name()
            # Event
            self.context_menu(action_text=action_text, position=position)
            # Evaluate
            current_color = self.plot_channel_a.color.name()
            self.assertNotEqual(previous_color, current_color)

        with self.subTest("2ChannelsSelected"):
            self.mouseClick_WidgetItem(self.plot_channel_b)
            self.plot_channel_b.setSelected(True)
            self.plot_channel_c.setSelected(True)
            position = self.plot.channel_selection.visualItemRect(self.plot_channel_c).center()
            # Setup
            previous_b_color = self.plot_channel_b.color.name()
            previous_c_color = self.plot_channel_c.color.name()
            # Event
            self.context_menu(action_text=action_text, position=position)
            # Evaluate
            current_b_color = self.plot_channel_b.color.name()
            current_c_color = self.plot_channel_c.color.name()
            self.assertNotEqual(previous_b_color, current_b_color)
            self.assertNotEqual(previous_c_color, current_c_color)
            self.assertNotEqual(current_b_color, current_c_color)

    def test_Action_CopyDisplayProperties_Group(self):
        """
        Test Scope:
            - Ensure that first channel display properties are copied in clipboard if item selected is group item.
        Events:
            - Insert Empty Group
            - Select Group
            - Trigger 'Copy display properties' action
            - Add items to group
            - Open Context Menu
            - Trigger 'Copy display properties' action
            - Remove channel from group
            - Open Context Menu
            - Trigger 'Copy display properties' action
        Evaluate:
            - Evaluate that channel display properties are stored in clipboard in json format.
        """
        with mock.patch("asammdf.gui.widgets.tree.QtWidgets.QInputDialog.getText") as mo_getText:
            mo_getText.return_value = "A", True
            self.context_menu(action_text="Add channel group")
            mo_getText.return_value = "B", True
            self.context_menu(action_text="Add channel group")
            mo_getText.return_value = "C", True
            self.context_menu(action_text="Add channel group")

        group_channel_a = self.plot.channel_selection.findItems("A", QtCore.Qt.MatchFlags())[0]
        group_channel_b = self.plot.channel_selection.findItems("B", QtCore.Qt.MatchFlags())[0]
        group_channel_c = self.plot.channel_selection.findItems("C", QtCore.Qt.MatchFlags())[0]

        QtWidgets.QApplication.instance().clipboard().clear()
        with self.subTest("EmptyGroup_0"):
            position = self.plot.channel_selection.visualItemRect(group_channel_a).center()
            self.context_menu(action_text="Copy display properties", position=position)
            clipboard = QtWidgets.QApplication.instance().clipboard().text()
            try:
                content = json.loads(clipboard)
            except JSONDecodeError:
                self.fail("Clipboard Content cannot be decoded as JSON content.")
            else:
                self.assertIsInstance(content, dict)
                self.assertTrue(content["type"] == "group")

        with self.subTest("PopulatedGroup"):
            # Add Channels to Group
            self.move_item_inside_channels_tree_widget(src=self.plot_channel_a, dst=group_channel_a)

            position = self.plot.channel_selection.visualItemRect(self.plot_channel_a).center()
            self.context_menu(action_text="Copy display properties", position=position)

            clipboard = QtWidgets.QApplication.instance().clipboard().text()
            try:
                content = json.loads(clipboard)
            except JSONDecodeError:
                self.fail("Clipboard Content cannot be decoded as JSON content.")
            else:
                self.assertIsInstance(content, dict)
                self.assertTrue(content["type"] == "channel")

        with self.subTest("EmptyGroup_1"):
            group_channel_a.removeChild(self.plot_channel_a)
            position = self.plot.channel_selection.visualItemRect(group_channel_a).center()
            self.context_menu(action_text="Copy display properties", position=position)
            clipboard = QtWidgets.QApplication.instance().clipboard().text()
            try:
                content = json.loads(clipboard)
            except JSONDecodeError:
                self.fail("Clipboard Content cannot be decoded as JSON content.")
            else:
                self.assertIsInstance(content, dict)
                self.assertTrue(content["type"] == "group")

        with self.subTest("EmptyGroup_2"):
            # Add Channels to Group
            self.move_item_inside_channels_tree_widget(src=group_channel_c, dst=group_channel_b)
            group_channel_a.setExpanded(True)
            group_channel_b.setExpanded(True)
            group_channel_c.setExpanded(True)
            self.move_item_inside_channels_tree_widget(src=group_channel_b, dst=group_channel_a)
            group_channel_a.setExpanded(True)
            group_channel_b.setExpanded(True)
            group_channel_c.setExpanded(True)
            position = self.plot.channel_selection.visualItemRect(group_channel_a).center()
            self.context_menu(action_text="Copy display properties", position=position)
            clipboard = QtWidgets.QApplication.instance().clipboard().text()
            try:
                content = json.loads(clipboard)
            except JSONDecodeError:
                self.fail("Clipboard Content cannot be decoded as JSON content.")
            else:
                self.assertIsInstance(content, dict)
                self.assertTrue(content["type"] == "group")

    def test_Action_PasteDisplayProperties_Group(self):
        """
        Test Scope:
            - Ensure that channel display properties can be pasted over different channel.
        Events:
            - Select one channel
            - Open Context Menu
            - Trigger 'Copy display properties' action
            - Select a group channel
            - Open Context Menu
            - Trigger 'Paste display properties' action
            - Select group channel
            - Open Context Menu
            - Trigger 'Copy display properties' action
            - Select a channel
            - Open Context Menu
            - Trigger 'Paste display properties' action
        Evaluate:
            - Evaluate that display properties are transferred from one channel to another
        """
        action_copy = "Copy display properties"
        action_paste = "Paste display properties"

        # Insert Group
        with mock.patch("asammdf.gui.widgets.tree.QtWidgets.QInputDialog.getText") as mo_getText:
            mo_getText.return_value = "A", True
            self.context_menu(action_text="Add channel group")
            mo_getText.return_value = "B", True
            self.context_menu(action_text="Add channel group")
            mo_getText.return_value = "C", True
            self.context_menu(action_text="Add channel group")

        # Add Channels to Group
        group_channel_a = self.plot.channel_selection.findItems("A", QtCore.Qt.MatchFlags())[0]
        group_channel_a.setExpanded(True)
        group_channel_c = self.plot.channel_selection.findItems("C", QtCore.Qt.MatchFlags())[0]
        group_channel_c.setExpanded(True)
        self.move_item_inside_channels_tree_widget(src=group_channel_c, dst=group_channel_a)
        group_channel_a.setExpanded(True)
        group_channel_c.setExpanded(True)
        self.move_item_inside_channels_tree_widget(src=self.plot_channel_b, dst=group_channel_c)

        # Add Channels to Group
        group_channel_b = self.plot.channel_selection.findItems("B", QtCore.Qt.MatchFlags())[0]
        group_channel_b.setExpanded(True)
        self.move_item_inside_channels_tree_widget(src=self.plot_channel_d, dst=group_channel_b)

        with self.subTest("FromChannel_ToGroup"):
            position_src = self.plot.channel_selection.visualItemRect(self.plot_channel_a).center()
            self.context_menu(action_text=action_copy, position=position_src)

            channel_a_properties = QtWidgets.QApplication.instance().clipboard().text()
            channel_a_properties = json.loads(channel_a_properties)

            # Paste
            position_dst = self.plot.channel_selection.visualItemRect(group_channel_a).center()
            self.context_menu(action_text=action_paste, position=position_dst)

            # Copy
            position_src = self.plot.channel_selection.visualItemRect(group_channel_a).center()
            self.context_menu(action_text=action_copy, position=position_src)

            group_channel_properties = QtWidgets.QApplication.instance().clipboard().text()
            group_channel_properties = json.loads(group_channel_properties)

            # Evaluate
            self.assertEqual(channel_a_properties["ranges"], group_channel_properties["ranges"])

        with self.subTest("FromGroup_ToChannel"):
            position_src = self.plot.channel_selection.visualItemRect(group_channel_a).center()
            self.context_menu(action_text=action_copy, position=position_src)

            # Paste
            position_dst = self.plot.channel_selection.visualItemRect(self.plot_channel_c).center()
            self.context_menu(action_text=action_paste, position=position_dst)

            # Copy
            position_src = self.plot.channel_selection.visualItemRect(self.plot_channel_c).center()
            self.context_menu(action_text=action_copy, position=position_src)

            channel_c_properties = QtWidgets.QApplication.instance().clipboard().text()
            channel_c_properties = json.loads(channel_c_properties)

            # Evaluate
            self.assertEqual(channel_c_properties["ranges"], group_channel_properties["ranges"])

        with self.subTest("FromGroup_ToGroup"):
            self.move_item_inside_channels_tree_widget(src=self.plot_channel_a, dst=group_channel_a)

            position_src = self.plot.channel_selection.visualItemRect(group_channel_b).center()
            self.context_menu(action_text=action_copy, position=position_src)
            group_channel_b_properties = QtWidgets.QApplication.instance().clipboard().text()
            group_channel_b_properties = json.loads(group_channel_b_properties)

            # Paste
            position_dst = self.plot.channel_selection.visualItemRect(group_channel_a).center()
            self.context_menu(action_text=action_paste, position=position_dst)

            # Copy
            position_src = self.plot.channel_selection.visualItemRect(group_channel_a).center()
            self.context_menu(action_text=action_copy, position=position_src)

            group_channel_a_properties = QtWidgets.QApplication.instance().clipboard().text()
            group_channel_a_properties = json.loads(group_channel_a_properties)

            # Evaluate
            self.assertEqual(group_channel_a_properties, group_channel_b_properties)

    def test_Action_CopyChannelStructure_Group(self):
        """
        Test Scope:
            - Ensure that channel structure is copied in clipboard.
        Events:
            - Select one group
            - Open Context Menu
            - Trigger 'Copy channel structure' action
        Evaluate:
            - Evaluate that group channel structure is stored in clipboard in json format.
        """
        with mock.patch("asammdf.gui.widgets.tree.QtWidgets.QInputDialog.getText") as mo_getText:
            mo_getText.return_value = "FirstGroup", True
            self.context_menu(action_text="Add channel group")

        # Add Channels to Group
        group_channel = self.plot.channel_selection.findItems("FirstGroup", QtCore.Qt.MatchFlags())[0]
        self.move_item_inside_channels_tree_widget(src=self.plot_channel_a, dst=group_channel)

        position = self.plot.channel_selection.visualItemRect(self.plot_channel_a).center()
        self.context_menu(action_text="Copy channel structure", position=position)

        clipboard = QtWidgets.QApplication.instance().clipboard().text()
        try:
            content = json.loads(clipboard)
        except JSONDecodeError:
            self.fail("Clipboard Content cannot be decoded as JSON content.")
        else:
            self.assertIsInstance(content, list)
            for channel_properties in content:
                self.assertIsInstance(channel_properties, dict)

    def test_Action_PasteChannelStructure_Group(self):
        """
        Test Scope:
            - Ensure that channel and structure is duplicated and structure is kept.
        Events:
            - Create group
            - Add channels to group
            - Open Context Menu
            - Trigger 'Copy channel structure' action
            - Trigger 'Paste channel structure' action over new group
        Evaluate:
            - Evaluate that group channel is duplicated and structure is kept.
        """
        action_copy = "Copy channel structure"
        action_paste = "Paste channel structure"

        with mock.patch("asammdf.gui.widgets.tree.QtWidgets.QInputDialog.getText") as mo_getText:
            mo_getText.return_value = "FirstGroup", True
            self.context_menu(action_text="Add channel group")

        # Add Channels to Group
        group_channel = self.plot.channel_selection.findItems("FirstGroup", QtCore.Qt.MatchFlags())[0]
        group_channel.setExpanded(True)
        self.move_item_inside_channels_tree_widget(src=self.plot_channel_c, dst=group_channel)
        self.move_item_inside_channels_tree_widget(src=self.plot_channel_d, dst=group_channel)

        # Copy Channel Structure
        position_src = self.plot.channel_selection.visualItemRect(group_channel).center()
        self.context_menu(action_text=action_copy, position=position_src)

        channels_count = self.plot.channel_selection.topLevelItemCount()

        # Paste Channel Structure
        position_src = self.plot.channel_selection.visualItemRect(self.plot_channel_a).center()
        self.context_menu(action_text=action_paste, position=position_src)
        self.processEvents(0.1)

        self.assertEqual(channels_count + 1, self.plot.channel_selection.topLevelItemCount())
        self.assertEqual(2, len(self.plot.channel_selection.findItems("FirstGroup", QtCore.Qt.MatchFlags())))

    def test_Menu_EnableDisable_Action_DisableSelected(self):
        """
        Test Scope:
            - Ensure that channel selected is disabled.
        Events:
            - Disable all
            - Open Context Menu
            - Select action "Disable selected" from sub-menu "Enable/disable"
            - Select two channel items
            - Open Context Menu
            - Select action "Disable selected" from sub-menu "Enable/disable"
            - Select channel group
            - Open Context Menu
            - Select action "Disable selected" from sub-menu "Enable/disable"
        Expected:
            - Evaluate that selected items were disabled.
        """
        with mock.patch("asammdf.gui.widgets.tree.QtWidgets.QInputDialog.getText") as mo_getText:
            mo_getText.return_value = "A", True
            self.context_menu(action_text="Add channel group")

        # Add Channels to Group
        group_channel = self.plot.channel_selection.findItems("A", QtCore.Qt.MatchFlags())[0]
        self.move_item_inside_channels_tree_widget(src=self.plot_channel_c, dst=group_channel)

        self.context_menu(action_text="Disable selected")

        # Ensure that all items are disabled
        count = self.plot.channel_selection.topLevelItemCount()
        for i in range(count):
            item = self.plot.channel_selection.topLevelItem(i)
            if item.type() != item.Info:
                self.assertEqual(QtCore.Qt.CheckState.Checked, item.checkState(self.Column.NAME))

        # Select two channels
        self.plot_channel_a.setSelected(True)
        self.plot_channel_b.setSelected(True)
        position_src = self.plot.channel_selection.visualItemRect(self.plot_channel_b).center()
        self.context_menu(action_text="Disable selected", position=position_src)

        self.assertEqual(QtCore.Qt.CheckState.Unchecked, self.plot_channel_a.checkState(self.Column.NAME))
        self.assertEqual(QtCore.Qt.CheckState.Unchecked, self.plot_channel_b.checkState(self.Column.NAME))

        # Select group
        position_src = self.plot.channel_selection.visualItemRect(group_channel).center()
        self.context_menu(action_text="Disable selected", position=position_src)

        self.assertEqual(QtCore.Qt.CheckState.Unchecked, group_channel.checkState(self.Column.NAME))

    def test_Menu_EnableDisable_Action_DeactivateGroups_Group(self):
        """
        Test Scope:
            - Ensure that group items can be disabled from Context Menu.
        Events:
            - Add two groups
            - Add channels to group
            - Add group in group
            - Select 1 group
            - Open Context Menu
            - Trigger "Enable/Disable -> Deactivate Groups" action
        Expected:
            - Group Items should be disabled
            - Sub-groups should be disabled
        """
        with mock.patch("asammdf.gui.widgets.tree.QtWidgets.QInputDialog.getText") as mo_getText:
            mo_getText.return_value = "Group_A", True
            self.context_menu(action_text="Add channel group")
            mo_getText.return_value = "Group_B", True
            self.context_menu(action_text="Add channel group")

        # Add Channels to Group
        group_a_channel = self.plot.channel_selection.findItems("Group_A", QtCore.Qt.MatchFlags())[0]
        group_a_channel.setExpanded(True)
        self.move_item_inside_channels_tree_widget(src=self.plot_channel_a, dst=group_a_channel)
        self.move_item_inside_channels_tree_widget(src=self.plot_channel_b, dst=group_a_channel)

        # Add Channels to Group
        group_b_channel = self.plot.channel_selection.findItems("Group_B", QtCore.Qt.MatchFlags())[0]
        group_b_channel.setExpanded(True)
        self.move_item_inside_channels_tree_widget(src=self.plot_channel_c, dst=group_b_channel)
        self.move_item_inside_channels_tree_widget(src=self.plot_channel_d, dst=group_b_channel)

        self.move_item_inside_channels_tree_widget(src=group_b_channel, dst=group_a_channel)
        group_a_channel.setExpanded(True)
        group_b_channel.setExpanded(True)

        position = self.plot.channel_selection.visualItemRect(group_b_channel).center()
        self.context_menu(action_text="Deactivate groups", position=position)

        # Evaluate
        self.assertEqual(False, group_a_channel.isDisabled())
        self.assertEqual(True, group_b_channel.isDisabled())
        for child_index in range(group_b_channel.childCount()):
            child = group_b_channel.child(child_index)
            self.assertEqual(True, child.isDisabled())
        for child_index in range(group_a_channel.childCount()):
            child = group_a_channel.child(child_index)
            if child.type() == child.Channel:
                self.assertEqual(False, child.isDisabled())

        position = self.plot.channel_selection.visualItemRect(group_a_channel).center()
        self.context_menu(action_text="Deactivate groups", position=position)

        # Evaluate
        self.assertEqual(True, group_a_channel.isDisabled())
        self.assertEqual(True, group_b_channel.isDisabled())
        for child_index in range(group_b_channel.childCount()):
            child = group_b_channel.child(child_index)
            self.assertEqual(True, child.isDisabled())
        for child_index in range(group_a_channel.childCount()):
            child = group_a_channel.child(child_index)
            self.assertEqual(True, child.isDisabled())

    def test_Menu_EnableDisable_Action_DisableAll(self):
        """
        Test Scope:
            - Ensure that all channels are disabled.
        Events:
            - Open Context Menu
            - Select action "Disable All" from sub-menu "Enable/disable"
            - Disable elements
            - Open Context Menu
            - Select action "Disable All" from sub-menu "Enable/disable"
        Expected:
            - Evaluate that all items are disabled.
        """
        with mock.patch("asammdf.gui.widgets.tree.QtWidgets.QInputDialog.getText") as mo_getText:
            mo_getText.return_value = "A", True
            self.context_menu(action_text="Add channel group")

        # Add Channels to Group
        group_channel = self.plot.channel_selection.findItems("A", QtCore.Qt.MatchFlags())[0]
        self.move_item_inside_channels_tree_widget(src=self.plot_channel_a, dst=group_channel)
        self.processEvents(1)

        # Ensure that all items are enabled
        count = self.plot.channel_selection.topLevelItemCount()
        for i in range(count):
            item = self.plot.channel_selection.topLevelItem(i)
            if item.type() != item.Info:
                self.assertEqual(QtCore.Qt.CheckState.Checked, item.checkState(self.Column.NAME))

        self.context_menu(action_text="Disable all")

        # Ensure that all items are disabled
        count = self.plot.channel_selection.topLevelItemCount()
        for i in range(count):
            item = self.plot.channel_selection.topLevelItem(i)
            if item.type() != item.Info:
                self.assertEqual(QtCore.Qt.CheckState.Unchecked, item.checkState(self.Column.NAME))
                # Disable items
                item.setCheckState(self.Column.NAME, QtCore.Qt.CheckState.Checked)

        self.context_menu(action_text="Disable all")

        for i in range(count):
            item = self.plot.channel_selection.topLevelItem(i)
            if item.type() != item.Info:
                self.assertEqual(QtCore.Qt.CheckState.Unchecked, item.checkState(self.Column.NAME))

    def test_Menu_EnableDisable_Action_EnableSelected(self):
        """
        Test Scope:
            - Ensure that channel selected is enabled.
        Events:
            - Disable all
            - Open Context Menu
            - Select action "Enable selected" from sub-menu "Enable/disable"
            - Select two channel items
            - Open Context Menu
            - Select action "Enable selected" from sub-menu "Enable/disable"
            - Select channel group
            - Open Context Menu
            - Select action "Enable selected" from sub-menu "Enable/disable"
        Expected:
            - Evaluate that selected items were enabled.
        """
        with mock.patch("asammdf.gui.widgets.tree.QtWidgets.QInputDialog.getText") as mo_getText:
            mo_getText.return_value = "A", True
            self.context_menu(action_text="Add channel group")

        # Add Channels to Group
        group_channel = self.plot.channel_selection.findItems("A", QtCore.Qt.MatchFlags())[0]
        self.move_item_inside_channels_tree_widget(src=self.plot_channel_c, dst=group_channel)

        # Disable all items
        count = self.plot.channel_selection.topLevelItemCount()
        for i in range(count):
            item = self.plot.channel_selection.topLevelItem(i)
            if item.type() != item.Info:
                item.setCheckState(self.Column.NAME, QtCore.Qt.CheckState.Unchecked)

        self.context_menu(action_text="Enable selected")

        # Ensure that all items are disabled
        count = self.plot.channel_selection.topLevelItemCount()
        for i in range(count):
            item = self.plot.channel_selection.topLevelItem(i)
            if item.type() != item.Info:
                self.assertEqual(QtCore.Qt.CheckState.Unchecked, item.checkState(self.Column.NAME))

        # Select two channels
        self.plot_channel_a.setSelected(True)
        self.plot_channel_b.setSelected(True)
        position_src = self.plot.channel_selection.visualItemRect(self.plot_channel_b).center()
        self.context_menu(action_text="Enable selected", position=position_src)

        self.assertEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_a.checkState(self.Column.NAME))
        self.assertEqual(QtCore.Qt.CheckState.Checked, self.plot_channel_b.checkState(self.Column.NAME))

        # Select group
        position_src = self.plot.channel_selection.visualItemRect(group_channel).center()
        self.context_menu(action_text="Enable selected", position=position_src)

        self.assertEqual(QtCore.Qt.CheckState.Checked, group_channel.checkState(self.Column.NAME))

    def test_Menu_EnableDisable_Action_EnableAll(self):
        """
        Test Scope:
            - Ensure that all channels are enabled.
        Events:
            - Open Context Menu
            - Select action "Enable All" from sub-menu "Enable/disable"
            - Disable elements
            - Open Context Menu
            - Select action "Enable All" from sub-menu "Enable/disable"
        Expected:
            - Evaluate that all items are enabled.
        """
        with mock.patch("asammdf.gui.widgets.tree.QtWidgets.QInputDialog.getText") as mo_getText:
            mo_getText.return_value = "A", True
            self.context_menu(action_text="Add channel group")

        # Add Channels to Group
        group_channel = self.plot.channel_selection.findItems("A", QtCore.Qt.MatchFlags())[0]
        self.move_item_inside_channels_tree_widget(src=self.plot_channel_a, dst=group_channel)

        # Ensure that all items are enabled:
        count = self.plot.channel_selection.topLevelItemCount()
        for i in range(count):
            item = self.plot.channel_selection.topLevelItem(i)
            if item.type() != item.Info:
                self.assertEqual(QtCore.Qt.CheckState.Checked, item.checkState(self.Column.NAME))

        self.context_menu(action_text="Enable all")

        # Ensure that all items are enabled:
        count = self.plot.channel_selection.topLevelItemCount()
        for i in range(count):
            item = self.plot.channel_selection.topLevelItem(i)
            if item.type() != item.Info:
                self.assertEqual(QtCore.Qt.CheckState.Checked, item.checkState(self.Column.NAME))
                # Disable items
                item.setCheckState(self.Column.NAME, QtCore.Qt.CheckState.Unchecked)

        self.context_menu(action_text="Enable all")

        for i in range(count):
            item = self.plot.channel_selection.topLevelItem(i)
            if item.type() != item.Info:
                self.assertEqual(QtCore.Qt.CheckState.Checked, item.checkState(self.Column.NAME))



from PySide6 import QtCore, QtGui, QtWidgets


class SearchTreeItem(QtWidgets.QTreeWidgetItem):
    def __init__(self, name, group, index, unit, source_name, source_path, comment):
        super().__init__([name, group, index, unit, source_name, source_path, comment])

        self.name = name
        self.group = group
        self.index = index
        self.unit = unit
        self.source_name = source_name
        self.source_path = source_path
        self.comment = comment

    def __del__(self):
        self.name = None
        self.group = None
        self.index = None
        self.unit = None
        self.source_name = None
        self.source_path = None
        self.comment = None


class SearchTreeWidget(QtWidgets.QTreeWidget):
    def __init__(self, can_delete_items=False, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setSelectionMode(QtWidgets.QAbstractItemView.SelectionMode.ExtendedSelection)
        self.setDragDropMode(QtWidgets.QAbstractItemView.DragDropMode.NoDragDrop)
        self.setUniformRowHeights(True)

        self.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self.open_menu)

        self.can_delete_items = can_delete_items

    def keyPressEvent(self, event):
        if event.key() == QtCore.Qt.Key.Key_Delete and self.can_delete_items:
            selected_items = self.selectedItems()

            root = self.invisibleRootItem()
            for item in selected_items:
                (item.parent() or root).removeChild(item)

            event.accept()
        else:
            super().keyPressEvent(event)

    def open_menu(self, position):
        count = 0
        iterator = QtWidgets.QTreeWidgetItemIterator(self)
        while iterator.value():
            iterator += 1
            count += 1

        self.context_menu = menu = QtWidgets.QMenu()
        menu.addAction(f"{count} items")
        menu.addSeparator()
        action = QtGui.QAction("Expand all", menu)
        action.triggered.connect(self.expandAll)
        menu.addAction(action)
        action = QtGui.QAction("Collapse all", menu)
        action.triggered.connect(self.collapseAll)
        menu.addAction(action)

        menu.exec(self.viewport().mapToGlobal(position))
